# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from dbmanagr.writer import FormatWriter
from dbmanagr.formatter import Formatter, DefaultFormatter


class ArgumentWriter(FormatWriter):
    def __init__(self, options=None):
        FormatWriter.__init__(self, u'{0}\n', u'{0}\n')
        self.options = options
        Formatter.set(DefaultFormatter())

    def str(self, items):
        options = self.options
        (includes, excludes, substitutes) = items
        output = []
        if options.includes and len(includes) > 0:
            output.append(u'-i {0}'.format(','.join(includes)))
        if options.excludes and len(excludes) > 0:
            output.append(u'-x {0}'.format(','.join(excludes)))
        if options.substitutes and len(substitutes) > 0:
            output.append(u'-s {0}'.format(','.join(substitutes)))
        return self.items_format.format(' '.join(output))


class ArgumentVerboseWriter(FormatWriter):
    def __init__(self, options=None):
        FormatWriter.__init__(self, u'{0}\n', u'{0}')
        if options.verbose > 0:
            Formatter.set(DefaultFormatter())
        else:
            Formatter.set(DefaultFormatter())

    def str(self, items):
        (includes, excludes, substitutes) = items
        output = []
        if len(includes) > 0:
            output.append('Includes:\n{0}\n\n'.format('\n'.join(
                map(lambda x: ' - {0}'.format(x), includes))))
        if len(excludes) > 0:
            output.append('Excludes:\n{0}\n\n'.format('\n'.join(
                map(lambda x: ' - {0}'.format(x), excludes))))
        if len(substitutes) > 0:
            output.append('Substitutes:\n{0}'.format('\n'.join(
                map(lambda x: ' - {0}'.format(x), substitutes))))
        return self.items_format.format(''.join(output))


class ArgumentTestWriter(ArgumentWriter):
    pass



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

from dbmanagr.logger import LogWith
from dbmanagr.utils import create_title

logger = logging.getLogger(__name__)


class Comment(object):
    def __init__(
            self, id_, title, subtitle, order, search, display, columns,
            aliases):
        self.id = id_
        self.title = title
        self.subtitle = subtitle
        self.order = order
        self.search = search
        self.display = display
        self.columns = columns
        self.aliases = aliases

    def __repr__(self):
        return unicode(self.__dict__)


@LogWith(logger)
def update_aliases(tablename, counter, aliases, foreign_keys):
    for key in filter(
            lambda k: foreign_keys[k].a.table.name == tablename,
            foreign_keys.keys()):
        fk = foreign_keys[key]
        fktable = fk.b.table
        if fktable.name in aliases:
            # We already know this alias
            continue
        aliases[fktable.name] = create_alias(fktable.name, counter)
    return aliases


@LogWith(logger)
def create_alias(tablename, counter):
    if tablename in counter:
        alias_format = u'_{0}{1}'
    else:
        alias_format = u'_{0}'
    counter[tablename] += 1
    return alias_format.format(tablename, counter[tablename])


@LogWith(logger)
def column_aliases(columns, alias):
    return dict(map(
        lambda col: (col.name, u'{{{0}_{1}}}'.format(alias, col.name)),
        columns))


def find_primary_key(table):
    if table.primary_key is None:
        # finds the primary key
        for c in table.columns():
            if c.primary_key:
                table.primary_key = c.name
                break

    return table.primary_key


def find_id(comment, caliases, alias, primary_key):
    if comment.id:
        return comment.id.format(**caliases)
    else:
        if primary_key:
            return u'{{{0}_{1}}}'.format(alias, primary_key)
        else:
            return "-"


@LogWith(logger)
def create_comment(table, comment, counter, aliases, alias):
    columns = {}
    display = []
    search = set(comment.search)

    logger.debug('search = set(%s)', comment.search)
    display.extend(comment.display)

    if table.name not in aliases:
        if alias:
            aliases[table.name] = alias
        else:
            aliases[table.name] = u'_{}'.format(table.name)

    alias = aliases[table.name]

    aliases = update_aliases(
        table.name,
        counter,
        aliases,
        table.foreign_keys())

    logger.debug('Aliases: %s', aliases)

    caliases = column_aliases(table.columns(), alias)
    for (_, v) in filter(
            lambda (k, v): v.a.table.name == table.name,
            table.foreign_keys().iteritems()):
        caliases.update(filter(
            lambda (k, v): k not in caliases.keys(),
            column_aliases(
                v.b.table.columns(), aliases[v.b.table.name]).iteritems()))

    logger.debug('Column aliases: %s', caliases)

    primary_key = find_primary_key(table)
    id_ = find_id(comment, caliases, alias, primary_key)

    title, subtitle, name = None, None, None
    if comment.title:
        title = comment.title.format(**caliases)
    else:
        name, title = create_title(comment, table.columns())
        d = dict(map(lambda k: (k.name, k.name), table.columns()))
        logger.debug('search.add(title.format(**d)=%s)', title.format(**d))
        search.add(title.format(**d))

        title = title.format(**caliases)

    if not subtitle:
        if comment.subtitle:
            subtitle = comment.subtitle.format(**caliases)
        else:
            _, subtitle = create_title(comment, table.columns(), [name])
            d = dict(map(lambda k: (k.name, k.name), table.columns()))
            logger.debug(
                'search.add(subtitle.format(**d)=%s)', subtitle.format(**d))
            search.add(subtitle.format(**d))

            subtitle = u'{0} (id={1})'.format(subtitle.format(**caliases), id_)

    if comment.order:
        order = comment.order
    else:
        order = []

    if display:
        display = map(lambda d: u'{0}_{1}'.format(alias, d), display)
    else:
        for column in table.columns():
            display.append(u'{0}_{1}'.format(alias, column.name))

    if primary_key in [c.name for c in table.columns()]:
        columns[primary_key] = id_
    else:
        columns[primary_key] = "'-'"

    if title != '*':
        columns['title'] = title

    columns['subtitle'] = subtitle
    for column in display:
        columns[column] = column

    return Comment(
        id_, title, subtitle, order, search, display, columns, aliases)



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from dbmanagr.args import parent_parser, format_group, create_parser

from .writer import GraphWriter, GraphvizWriter, GraphTestWriter, YamlWriter

parent = parent_parser(daemonable=True)

group = format_group(
    parent,
    GraphTestWriter)
group.add_argument(
    '-D',
    '--default',
    default=True,
    help='output format: human readable hierarchical text',
    dest='formatter',
    action='store_const',
    const=GraphWriter)
group.add_argument(
    '-G',
    '--graphviz',
    help='output format: a Graphviz graph',
    dest='formatter',
    action='store_const',
    const=GraphvizWriter)
group.add_argument(
    '-Y',
    '--yaml',
    help='output format: the graph in YAML format',
    dest='formatter',
    action='store_const',
    const=YamlWriter)

parser = create_parser(
    prog='dbgraph',
    description='A database visualisation tool that creates graphs from the '
                'database structure',
    parents=[parent])
parser.add_argument(
    'uri',
    help='the URI to parse (format for PostgreSQL/MySQL: '
         'user@host/database/table; for SQLite: databasefile.db/table)')
parser.add_argument(
    '-c',
    '--columns',
    dest='include_columns',
    default=False,
    help='include columns in output',
    action='store_true')
parser.add_argument(
    '-C',
    '--no-columns',
    dest='include_columns',
    default=True,
    help='don\'t include columns in output',
    action='store_false')
parser.add_argument(
    '--back-references',
    dest='include_back_references',
    default=True,
    help='include back references in output',
    action='store_true')
parser.add_argument(
    '--no-back-references',
    dest='include_back_references',
    default=False,
    help='don\'t include back references in output',
    action='store_false')
parser.add_argument(
    '--driver',
    dest='include_driver',
    default=False,
    help='include database driver in output (does not work well with graphviz '
         'as output)',
    action='store_true')
parser.add_argument(
    '--no-driver',
    dest='include_driver',
    default=True,
    help='don\'t include database driver in output',
    action='store_false')
parser.add_argument(
    '--connection',
    dest='include_connection',
    default=False,
    help='include connection in output (does not work well with graphviz as '
         'output)',
    action='store_true')
parser.add_argument(
    '--no-connection',
    dest='include_connection',
    default=True,
    help='don\'t include connection in output',
    action='store_false')
parser.add_argument(
    '--database',
    dest='include_database',
    default=False,
    help='include database in output (does not work well with graphviz as '
         'output)',
    action='store_true')
parser.add_argument(
    '--no-database',
    dest='include_database',
    default=True,
    help='don\'t include database in output',
    action='store_true')
parser.add_argument(
    '-M',
    '--max-depth',
    default=-1,
    type=int,
    help='the maximum depth to use in recursion/inclusion (default: '
         '%(default)s)')

group = parser.add_mutually_exclusive_group()
group.add_argument(
    '-r',
    '--recursive',
    help='include any forward/back reference to the starting table, recursing '
         'through all tables eventually',
    action='store_true')
group.add_argument(
    '-i',
    '--include',
    help='include the specified columns and their foreign rows, if any. '
         'Multiple columns can be specified by separating them with a comma '
         '(,)')
parser.add_argument(
    '-x',
    '--exclude',
    help='exclude the specified columns')
parser.add_argument(
    '-v',
    '--verbose',
    action='count',
    help='specify the verbosity of the output, increase the number of '
         'occurences of this option to increase verbosity')



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#


class Op(object):
    def last(self):
        return None

    def __str__(self):
        return self.__repr__()


class QueryFilter(Op):
    def __init__(self, lhs, operator=None, rhs=None):
        self.lhs = lhs
        self.operator = operator
        self.rhs = rhs

    def last(self):
        return self

    def __repr__(self):
        return '{me.lhs} {me.operator} {me.rhs}'.format(
            me=self)


class BitOp(Op):
    def __init__(self, children=None):
        if children is not None:
            self.children = children
        else:
            self.children = []

    def append(self, _filter):
        self.children.append(_filter)

    def last(self):
        return self.children[-1].last()

    def __getitem__(self, i):
        return self.children[i]

    def __len__(self):
        return len(self.children)


class OrOp(BitOp):
    def __repr__(self):
        return u' or '.join(map(unicode, self.children))


class AndOp(BitOp):
    def __repr__(self):
        return u' and '.join(map(unicode, self.children))



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from dbmanagr.writer import FormatWriter, StdoutWriter
from dbmanagr.formatter import Formatter, DefaultFormatter
import datetime

import pprint
pp = pprint.PrettyPrinter(indent=4)


def sql_escape(value):
    if type(value) is str or type(value) is unicode:
        return "'%s'" % value
    if type(value) in [datetime.datetime, datetime.date, datetime.time]:
        return "'%s'" % value
    if type(value) is bool:
        return unicode(value).lower()
    return unicode(value)


class ExecuteWriter(StdoutWriter):
    def __init__(self, options=None):
        StdoutWriter.__init__(self, u'{0}\n', u'{item}')
        Formatter.set(DefaultFormatter())
        self.options = options


class SqlInsertWriter(FormatWriter):
    def __init__(self, options=None):
        FormatWriter.__init__(
            self,
            u'{0}\n',
            u'insert into {table} ({columns}) values ({values});')
        Formatter.set(DefaultFormatter())
        self.table_name = options.table_name

    def itemtostring(self, item):
        row = item.row
        return self.item_format.format(
            table=self.table_name,
            columns=self.create_columns(row.keys()),
            values=self.create_values(row.values()))

    def create_columns(self, cols):
        return u','.join(map(unicode, cols))

    def create_values(self, values):
        return u','.join(map(sql_escape, values))


class ExecuteTestWriter(ExecuteWriter):
    pass



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from dbmanagr.args import parent_parser, create_parser

parent = parent_parser(daemonable=True, daemon=True)

parser = create_parser(
    prog='dbdaemon',
    description='The daemon background process.',
    parents=[parent])
parser.add_argument(
    'command',
    choices=['start', 'stop', 'restart', 'status'],
    help='the command to issue'
)



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from dbmanagr.args import parent_parser, format_group, create_parser
from dbmanagr.args import CommaSeparatedDict

from .writer import SqlInsertWriter, SqlUpdateWriter, SqlDeleteWriter
from .writer import YamlWriter, FormattedWriter

parent = parent_parser(daemonable=True)

group = format_group(
    parent,
    SqlInsertWriter)
group.add_argument(
    '-I',
    '--insert',
    default=True,
    help='output format: SQL insert statements',
    dest='formatter',
    action='store_const',
    const=SqlInsertWriter)
group.add_argument(
    '-U',
    '--update',
    help='output format: SQL update statements',
    dest='formatter',
    action='store_const',
    const=SqlUpdateWriter)
group.add_argument(
    '-D',
    '--delete',
    help='output format: SQL delete statements',
    dest='formatter',
    action='store_const',
    const=SqlDeleteWriter)
group.add_argument(
    '-Y',
    '--yaml',
    help='output format: YAML data',
    dest='formatter',
    action='store_const',
    const=YamlWriter)
group.add_argument(
    '-F',
    '--formatted',
    help='output format: given with the -f/--format option',
    dest='formatter',
    action='store_const',
    const=FormattedWriter)

parser = create_parser(
    prog='dbexport',
    description='An export tool that exports database rows in different '
                'formats.',
    parents=[parent])
parser.add_argument(
    'uri',
    help='the URI to parse (format for PostgreSQL/MySQL: '
         'user@host/database/table?column=value; for SQLite: '
         'databasefile.db/table?column=value)')
parser.add_argument(
    '-i',
    '--include',
    help='include specified columns; referenced rows, if any, can be included '
         'by adding a dot after the column name; multiple columns can be '
         'specified by separating them with a comma')
parser.add_argument(
    '-x',
    '--exclude',
    help='exclude specified columns')
parser.add_argument(
    '-s',
    '--substitutes',
    action=CommaSeparatedDict,
    help='substitute values of rows with static values, in the form of: '
         'fk.column=replacement; multiple substitutions can be specified by '
         'separating them with a comma')
parser.add_argument(
    '-m',
    '--limit',
    type=int,
    default=50,
    help='limit the results of the main query to this amount of rows '
         '(default: %(default)s)')
parser.add_argument(
    '-p',
    '--package',
    default='models',
    help='the package for YAML entities (default: %(default)s)')
parser.add_argument(
    '-f',
    '--format',
    default=u'{0}',
    help='the format for the -F/--formatted writer (use {0} for positional '
         'arguments, or {column_name} to insert the actual value of '
         'table.column_name) (default: %(default)s)')



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from sqlalchemy import Integer

from dbmanagr.logger import logger, LogWith
from dbmanagr.options import format_value, UriOptionsParser
from dbmanagr.driver import DatabaseDriver


@LogWith(logger)
def restriction(alias, column, operator, value, map_null_operator=True):
    if not column:
        raise Exception('Parameter column may not be None!')
    if operator in ['~', 'like'] and isinstance(column.type, Integer):
        try:
            int(value)
            # LIKE not allowed on integer columns, change operator to
            # equals
            operator = '='
        except ValueError:
            pass

    if alias:
        alias = '{0}.'.format(alias)
    else:
        alias = ''
    lhs = column.name
    if column.table:
        lhs = '{0}{1}'.format(alias, column.name)
    if (value
            and isinstance(column.type, Integer)
            and type(value) is not list):
        try:
            int(value)
        except ValueError:
            # column type is integer, but value is not
            lhs = 'cast({0}{1} as char)'.format(alias, column.name)
    if operator in ['=', '!='] and (value == 'null' or value is None):
        if map_null_operator:
            operator = {
                '=': 'is',
                '!=': 'is not'
            }.get(operator)
        value = None
    rhs = format_value(column, value)

    return ' '.join([lhs, operator, rhs])


class MySQLDriver(DatabaseDriver):
    def __init__(self):
        self.user = None
        self.host = None
        self.gen = None

    @LogWith(logger)
    def restriction(self, *args):
        return restriction(*args)

    def __repr__(self):
        return str(self.__dict__)


class MySQLOptionsParser(UriOptionsParser):
    def create_driver(self):
        return MySQLDriver()



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import json

COMMENT_ID = 'id'
COMMENT_TITLE = 'title'
COMMENT_SUBTITLE = 'subtitle'
COMMENT_ORDER_BY = 'order'
COMMENT_SEARCH = 'search'
COMMENT_DISPLAY = 'display'

logger = logging.getLogger(__name__)


class TableComment(object):
    """The comment on the given table that allows to display much more
accurate information"""

    def __init__(self, json_string):
        self.id = None
        self.title = None
        self.subtitle = None
        self.search = None
        self.display = None
        self.order = None

        self.parse(json_string)

    def parse(self, json_string):
        d = {
            COMMENT_ORDER_BY: [],
            COMMENT_SEARCH: [],
            COMMENT_DISPLAY: []
        }

        if json_string:
            try:
                d.update(json.loads(json_string))
            except BaseException as e:
                logger.warn('Error parsing JSON comment: %s', e)

        if COMMENT_ID in d:
            self.id = d[COMMENT_ID]
        if COMMENT_TITLE in d and self.id:
            d[COMMENT_TITLE] = u'{0}.%s' % self.id
        if COMMENT_TITLE in d:
            self.title = d[COMMENT_TITLE]
        if COMMENT_SUBTITLE in d:
            self.subtitle = d[COMMENT_SUBTITLE]
        self.search = d[COMMENT_SEARCH]
        self.display = d[COMMENT_DISPLAY]
        self.order = d[COMMENT_ORDER_BY]

    def __repr__(self):
        return self.__dict__.__repr__()



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

__all__ = ["databaseconnection", "driver"]

from os.path import expanduser
from os import getenv
from collections import OrderedDict

from dbmanagr import __drivers__
from dbmanagr.utils import module_installed
from dbmanagr.sources.source import Source
from dbmanagr.sources.anypass import AnyPassSource
from dbmanagr.sources.dbexplorer import DBExplorerSource
from dbmanagr.options import Options
from .databaseconnection import PostgreSQLConnection
from .driver import PostgreSQLOptionsParser

DRIVERS = OrderedDict([
    ('psycopg2', 'postgresql+psycopg2://{user}:{password}@{host}/{database}'),
    ('postgresql',
        'postgresql+pypostgresql://{user}:{password}@{host}/{database}'),
    ('pg8000', 'postgresql+pg8000://{user}:{password}@{host}/{database}'),
    # ('zxjdbc', 'postgresql+zxjdbc://{user}:{password}@{host}/{database}')
])


def init_postgresql(driver, dbexplorer_config, pgpass_config):
    Source.sources.append(
        DBExplorerSource(
            driver, dbexplorer_config, 'postgresql', PostgreSQLConnection))
    Source.sources.append(
        AnyPassSource(
            driver, pgpass_config, PostgreSQLConnection))
    # Doesn't make much sense at the moment - passwords are encrypted in the
    # plist file
    # Source.sources.append(NavicatPostgreSQLSource(navicat_config))


def init():
    module = module_installed(*DRIVERS.keys())
    if not module:
        return

    __drivers__.append(module)
    init_postgresql(
        DRIVERS[module],
        getenv(
            'DBEXPLORER_CFG',
            expanduser('~/.dbexplorer/dbexplorer.cfg')),
        getenv(
            'PGPASS_CFG',
            expanduser('~/.pgpass'))
    )

    Options.parser['postgresql'] = PostgreSQLOptionsParser()



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from .baseitem import BaseItem
from dbmanagr import KIND_FOREIGN_VALUE, KIND_FOREIGN_KEY, KIND_VALUE
from dbmanagr import IMAGE_FOREIGN_VALUE, IMAGE_FOREIGN_KEY, IMAGE_VALUE

TITLES = {
    KIND_FOREIGN_VALUE: u'← %s',
    KIND_FOREIGN_KEY: u'→ %s',
    KIND_VALUE: u'%s'
}
ICONS = {
    KIND_FOREIGN_VALUE: IMAGE_FOREIGN_VALUE,
    KIND_FOREIGN_KEY: IMAGE_FOREIGN_KEY,
    KIND_VALUE: IMAGE_VALUE
}


class Value(BaseItem):
    """A value from the database"""

    def __init__(self, value, subtitle, autocomplete, validity, kind):
        self._value = value
        self._subtitle = subtitle
        self._autocomplete = autocomplete
        self._validity = validity
        self._kind = kind

    def title(self):
        if type(self._value) is buffer:
            return '[BLOB]'
        return TITLES.get(self._kind, KIND_VALUE) % self._value

    def subtitle(self):
        return self._subtitle

    def autocomplete(self):
        return self._autocomplete

    def validity(self):
        return self._validity

    def icon(self):
        return ICONS.get(self._kind, KIND_VALUE)

    def value(self):
        return self._value

    def as_json(self):
        return {
            '__cls__': str(self.__class__),
            'value': self._value,
            'subtitle': self._subtitle
        }



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import argparse
import logging

from dbmanagr.version import __version__
from dbmanagr import __drivers__

PARSER_ARGS = {
    'formatter_class': argparse.RawDescriptionHelpFormatter,
    'epilog': """Contact:
  If you experience bugs or want to request new features please visit
  <https://github.com/resamsel/dbmanagr/issues>"""
}


class CommaSeparatedDict(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(
            namespace,
            self.dest,
            dict(map(
                lambda s: map(
                    lambda s: s.strip(),
                    (s.split('=', 1) + [''])[:2]),
                filter(len, values.split(','))))
        )


class CommaSeparatedStringList(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(
            namespace,
            self.dest,
            ','.join(map("'{0}'".format, values.split(',')))
        )


class CommaSeparatedPlainList(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(
            namespace,
            self.dest,
            ','.join(values.split(','))
        )


def default_log_file(dirs=None):
    if dirs is None:
        dirs = ['/var/log', '/usr/local/var/log', '/tmp']
    for d in dirs:
        if os.access(d, os.W_OK):
            return os.path.join(d, 'dbmanagr.log')
    return os.path.expanduser('~/dbmanagr.log')


def parent_parser(daemonable=False, daemon=False):
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        '--version',
        action='version',
        version='%(prog)s {version} (drivers: {drivers})'.format(
            version=__version__, drivers=', '.join(__drivers__)))
    group = parser.add_argument_group('logging')
    group.add_argument(
        '--trace',
        action='store_true',
        default=False,
        help='set loglevel to trace')
    group.add_argument(
        '--debug',
        dest='loglevel',
        action='store_const',
        const=logging.DEBUG,
        help='set loglevel to debug')
    group.add_argument(
        '--info',
        dest='loglevel',
        action='store_const',
        const=logging.INFO,
        help='set loglevel to info')
    group.add_argument(
        '--warning',
        dest='loglevel',
        action='store_const',
        const=logging.WARNING,
        help='set loglevel to warning')
    group.add_argument(
        '--error',
        dest='loglevel',
        action='store_const',
        const=logging.ERROR,
        help='set loglevel to error')
    group.add_argument(
        '--critical',
        dest='loglevel',
        action='store_const',
        const=logging.CRITICAL,
        help='set loglevel to critical')
    group.add_argument(
        '-L',
        '--logfile',
        type=argparse.FileType('a'),
        default=default_log_file(),
        help='the file to log to')
    if daemonable:
        group = parser.add_argument_group('daemon')
        if daemon:
            group.add_argument(
                '--host',
                default='',
                help='the host of the daemon')
            group.add_argument(
                '--port',
                default=8020,
                type=int,
                help='the port of the daemon')
        else:
            group.add_argument(
                '--daemon',
                action='store_true',
                default=False,
                help='use a background process to speed up requests')
            group.add_argument(
                '--daemon-host',
                dest='host',
                default='',
                help='the host of the daemon')
            group.add_argument(
                '--daemon-port',
                dest='port',
                default=8020,
                type=int,
                help='the port of the daemon')
    return parser


def format_group(parser, test_writer=None):
    group = parser.add_argument_group('formatters')
    if test_writer:
        group.add_argument(
            '-T',
            '--test',
            help='output format: test specific',
            dest='formatter',
            action='store_const',
            const=test_writer)
    return group


def create_parser(**kwargs):
    kwargs.update(PARSER_ARGS)
    return argparse.ArgumentParser(**kwargs)



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

from collections import Counter
from sqlalchemy import or_, and_, String
from sqlalchemy.orm import aliased
from sqlalchemy.orm.session import Session

from dbmanagr.logger import LogWith
from dbmanagr.utils import create_title, operation
from dbmanagr.comment import create_comment
from dbmanagr.exception import UnknownColumnException
from dbmanagr.queryfilter import QueryFilter, OrOp, BitOp

logger = logging.getLogger(__name__)


def column_or_raise(table, columnname):
    column = table.column(columnname)
    if not column:
        raise UnknownColumnException(table, 'id')
    return column


@LogWith(logger)
def allowed(column, operator, value):
    if operator in ['~', '*']:
        return isinstance(column.type, String)
    if operator in ['=', '!=']:
        try:
            column.type.python_type(value)
            return True
        except BaseException:
            return False
    return True


def add_references(tablename, foreign_keys, joins, comment):
    for _, fk in filter(
            lambda (k, v): (
                v.a.table.name == tablename
                and (comment is None or k in comment.display)),
            foreign_keys.iteritems()):
        fktable = fk.b.table
        fkentity = fktable.entity()

        # Prevent multiple joins of the same table
        add_join(fkentity, joins)

    return joins


@LogWith(logger)
def add_join(entity, joins):
    if entity.name not in joins.keys():
        joins[entity.name] = aliased(
            entity, name='_{0}'.format(entity.name))

    return joins


def replace_filter(f, table, entity, comment, search_fields):
    if isinstance(f, BitOp):
        f.children = map(
            lambda child: replace_filter(
                child, table, entity, comment, search_fields),
            f.children)
    elif f.lhs == '' and search_fields:
        logger.debug('Search fields: %s', search_fields)
        rhs = f.rhs
        if rhs == '' and f.operator == '*':
            rhs = '%'
        ors = OrOp(map(
            lambda s: QueryFilter(s, f.operator, rhs),
            filter(
                lambda s: s in entity.columns,
                search_fields)))
        if 'id' in comment.columns:
            col = column_or_raise(table, 'id')
            ors.append(QueryFilter(col.name, f.operator, rhs))
        logger.debug('Searches: %s', ors)
        return ors
    return f


@LogWith(logger)
def add_filters(f, filters, table, joins):
    if isinstance(f, BitOp):
        op = {
            'OrOp': or_,
            'AndOp': and_
        }.get(f.__class__.__name__)
        fs = []

        for c in f.children:
            add_filters(c, fs, table, joins)

        logger.debug('Filter fs: %s', fs)

        if len(fs) > 1:
            filters.append(op(*fs))
        elif len(fs) > 0:
            filters.append(*fs)

    elif isinstance(f, QueryFilter):
        logger.debug(
            'Filter: lhs=%s, op=%s, rhs=%s',
            f.lhs, f.operator, f.rhs)
        if f.lhs != '':
            add_filter(f, filters, table, joins)


@LogWith(logger)
def add_filter(f, filters, table, joins):
    if not f.operator:
        return None

    if '.' in f.lhs:
        foreign_keys = table.foreign_keys()
        ref, colname = f.lhs.split('.', 1)
        if ref not in foreign_keys:
            raise UnknownColumnException(table, ref)
        t = foreign_keys[ref].b.table
        add_join(t.entity(), joins)
        return add_filter(
            QueryFilter(colname, f.operator, f.rhs), filters, t, joins)

    colname = f.lhs
    col = table.column(colname)
    if not col:
        raise UnknownColumnException(table, colname)
    tentity = joins[table.name]
    if allowed(tentity.columns[col.name], f.operator, f.rhs):
        op = operation(
            tentity.columns[col.name],
            f.operator,
            f.rhs)
        filters.append(op)
        return op
    else:
        filters.append(False)

    return None


def create_label(alias_format):
    return lambda column: column.label(alias_format.format(col=column))


def simplify(table, comment, key, d):
    if comment:
        d[key] = unicode(comment.__dict__[key]).format(table.name, **d)
        d['id'] = d[u'{0}_{1}'.format(
            comment.aliases[table.name], table.primary_key)]
    else:
        d[key] = create_title(
            comment, table.columns())[1].format(table.name, **d)


def with_filters(query, filters):
    """Adds filters"""

    for f in filters:
        query = query.filter(f)

    return query


def with_orders(query, orders):
    """Adds orders"""

    for order in orders:
        query = query.order_by(order)

    return query


class SimplifyMapper(object):
    def __init__(self, table, comment=None):
        self.table = table
        self.comment = comment

    def map(self, row):
        d = row.__dict__
        for k in filter(lambda k: k not in d, ['title', 'subtitle']):
            simplify(self.table, self.comment, k, d)
        return row


class QueryBuilder(object):
    def __init__(
            self,
            connection,
            table,
            filter_=None,
            order=None,
            limit=None,
            simplify=True):
        self.connection = connection
        self.table = table
        self.filter = filter_ if filter_ else OrOp()
        self.order = order if order else []
        self.limit = limit
        self.aliases = {}
        self.counter = Counter()
        self.simplify = simplify

        self.alias = '_%s' % self.table.name

    @LogWith(logger)
    def build(self):
        foreign_keys = self.table.foreign_keys()
        search_fields = []

        entity = aliased(self.table.entity(), name=self.alias)

        projection = map(lambda x: x, entity.columns)
        joins = {self.table.name: entity}

        if self.simplify:
            # Add referenced tables from comment to be linked
            comment = create_comment(
                self.table,
                self.connection.comment(self.table.name),
                self.counter,
                self.aliases,
                self.alias)

            add_references(self.table.name, foreign_keys, joins, comment)

            logger.debug('Joins: %s', joins)

            if not self.order:
                self.order = [self.table.primary_key]

            logger.debug('Order: %s', self.order)

            keys = dict(map(lambda k: (str(k), k), comment.columns.keys()))
            if comment.search:
                for s in comment.search:
                    search_fields.append(s.format(**keys))

            if not search_fields:
                search_fields.append(comment.title.format(**keys))

            logger.debug('Search fields: %s', search_fields)

            replace_filter(
                self.filter, self.table, entity, comment, search_fields)

        logger.debug('Aliases: %s', self.aliases)

        filters = []
        if self.filter:
            add_filters(self.filter, filters, self.table, joins)

        orders = []
        if self.order:
            for order in self.order:
                orders.append(entity.columns[order])

        # Create a session
        session = Session(self.connection.engine())

        # Create query
        if self.simplify:
            alias_format = '{col.table.name}_{col.name}'
        else:
            alias_format = '{col.name}'
        logger.debug('Projection: %s', projection)
        query = session.query(*map(create_label(alias_format), projection))
        logger.debug('Query (init): %s', query)

        # Add found joins
        # Aliased joins
        joins = dict(filter(
            lambda (k, v): k != entity.original.name,
            joins.iteritems()))
        logger.debug('Joins: %s', joins)
        for _, join in joins.iteritems():
            query = query.outerjoin(join)
            for column in join.columns.keys():
                col = join.columns[column]
                query = query.add_column(create_label(alias_format)(col))
        logger.debug('Query (joins): %s', query)

        query = with_filters(query, filters)
        query = with_orders(query, orders)

        logger.debug('Slice: 0, %d', self.limit)

        # For Markus: don't slice if limit is less than 1!
        if self.limit > 0:
            return query.slice(0, self.limit)

        return query



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import sqlparse
import re

from datetime import datetime

from dbmanagr.wrapper import Wrapper
from dbmanagr.config import Config
from dbmanagr.writer import Writer
from dbmanagr.utils import find_connection
from dbmanagr.sources.source import Source
from dbmanagr.logger import logger, LogTimer, log_error
from dbmanagr.exception import UnknownConnectionException
from dbmanagr.dto.mapper import to_dto

from .args import parser
from .writer import ExecuteWriter

EXECUTION_START = {
    1: '{statement}\n',
    2: '{time:%H:%M:%S.%f}: executing: {statement}\n',
    3: '{time:%Y-%m-%d %H:%M:%S.%f}: executing: {statement}\n'
}
EXECUTION_END = {
    2: '{time:%H:%M:%S.%f}: execution took: {duration}s\n',
    3: '{time:%Y-%m-%d %H:%M:%S.%f}: execution took: {duration}s\n'
}


class Item(object):
    def __init__(self, connection, row):
        self.connection = connection
        self.row = row

    def __str__(self):
        return '\t'.join(map(unicode, self.row))


def read_sqls(files):
    sql = ''

    for file_ in files:
        sql += read_sql(file_)

    return sql


def read_sql(file_):
    timer = LogTimer(logger, 'Reading input statements')

    try:
        sql = file_.read()
    except KeyboardInterrupt:  # pragma: no cover
        sql = None
    finally:
        file_.close()

    timer.stop()

    return sql


def read_statements(opts):
    if opts.statements is not None:
        sql = opts.statements
    else:
        sql = read_sqls(opts.infile)

    if not sql:
        return None

    timer = LogTimer(logger, 'Splitting SQL statements')

    # Removes the shebang, if any
    sql = re.sub(r'^#!.*\n', '', sql)

    stmts = filter(lambda s: len(s.strip()) > 0, sqlparse.split(sql))

    timer.stop()

    logger.info('Number of SQL statements: %d', len(stmts))

    return stmts


def trim_space(stmt):
    return re.sub(r'\s+', ' ', unicode(stmt))


class BaseExecuter(object):
    def begin(self):  # pragma: no cover
        pass

    def commit(self):  # pragma: no cover
        pass

    def rollback(self):  # pragma: no cover
        pass

    def execute(self, stmt):  # pragma: no cover
        pass

    def write(self, items):
        sys.stdout.write(Writer.write(items))
        sys.stdout.write('\n')


class DefaultExecuter(BaseExecuter):
    """Execute in wrapped transaction"""

    def __init__(self, connection, opts):
        self.connection = connection
        self.opts = opts
        self.trans = None

    def begin(self):
        self.trans = self.connection.begin()

    def commit(self):
        if self.trans:
            self.trans.commit()

    def rollback(self):
        if self.trans:
            self.trans.rollback()

    def execute(self, stmt):
        changes = 0
        results = []

        result = self.connection.execute(stmt)
        if result.cursor:
            results = map(lambda row: Item(self.connection, row), result)
            self.write(results)
        else:
            # increase changes based on the returned result info
            changes += result.rowcount

        return (results, changes, 0)


class IsolationExecuter(BaseExecuter):
    """Execute in wrapped transaction"""

    def __init__(self, connection, opts):
        self.connection = connection
        self.opts = opts

    def execute(self, stmt):
        results = []
        changes = 0
        errors = 0

        try:
            # If we're about to ignore errors we need to
            # use a separate transaction for each
            # statement - otherwise previously successful
            # executions would get lost
            trans = self.connection.begin()

            result = self.connection.execute(stmt)
            if result.cursor:
                results = map(lambda row: Item(self.connection, row), result)
                self.write(results)
            else:
                # increase changes based on the returned result
                # info
                changes = result.rowcount

            if self.opts.dry_run:
                trans.rollback()
            else:
                trans.commit()
        except BaseException as e:
            trans.rollback()
            if not self.opts.mute_errors:
                log_error(e)
            errors += 1

        return (results, changes, errors)


class DatabaseExecuter(Wrapper):
    """The main class"""
    def __init__(self, options):
        Wrapper.__init__(self, options)

        if options.formatter:
            Writer.set(options.formatter(options))
        else:
            Writer.set(ExecuteWriter())

    def write(self):
        try:
            self.run()
        except BaseException:
            return -1
        return 0

    def execute(self):
        """The main method that splits the arguments and starts the magic"""
        options = self.options

        # search exact match of connection
        connection, opts = find_connection(
            Source.connections(),
            options,
            lambda con, opts: (
                con.matches(opts)
                and opts.show in ['databases', 'tables', 'columns', 'values']))

        if connection is None:
            raise UnknownConnectionException(
                options.uri,
                map(lambda c: c.autocomplete(), Source.connections()))

        return self.process(connection, opts)

    def process(self, connection, opts):
        # Reads the statements
        stmts = read_statements(opts)

        # Exit gracefully when no statements have been found (or the
        # input got cancelled)
        if not stmts:
            return []

        # Collects results
        results = []
        # Counts the changes (inserts, updates)
        changes = 0
        # Counts the errors (useful with option --ignore-errors)
        errors = 0
        # Counts the statements
        counter = 0

        timer = None
        executer = None
        try:
            # Connects to the database and starts a transaction
            connection.connect(opts.database)

            timer = LogTimer(logger, 'Executing SQL statements')

            if opts.isolate_statements:
                executer = IsolationExecuter(connection, opts)
            else:
                executer = DefaultExecuter(connection, opts)

            executer.begin()

            for stmt in stmts:
                start = datetime.now()
                sys.stdout.write(
                    EXECUTION_START.get(opts.verbose, '').format(
                        time=start,
                        statement=stmt))

                res, changed, failed = executer.execute(stmt)
                results.extend(res)
                changes += changed
                errors += failed
                counter += 1

                if opts.progress > 0 and counter % opts.progress == 0:
                    sys.stderr.write('.')
                    sys.stderr.flush()

                end = datetime.now()
                sys.stdout.write(
                    EXECUTION_END.get(opts.verbose, '').format(
                        time=end,
                        statement=trim_space(stmt),
                        duration=(end - start).total_seconds()))

            if opts.dry_run:
                executer.rollback()
            else:
                executer.commit()

            if opts.progress > 0 and counter >= opts.progress:
                # Write a new line after progress indicator dots have
                # been written
                sys.stderr.write('\n')
        except BaseException:
            errors += 1
            if executer:
                executer.rollback()
            if not opts.mute_errors:
                raise
        finally:
            connection.close()
            if timer:
                timer.stop()

        if not results:
            dry_run = ''
            if opts.dry_run:
                dry_run = ' (dry run)'
            sys.stdout.write(
                'Changed rows: {0}{1}\n'.format(changes, dry_run))
        if errors:
            sys.stdout.write('Errors: {0}'.format(errors))

        return to_dto(results)


def execute(args):
    """
    Directly calls the execute method and avoids using the wrapper
    """
    return DatabaseExecuter(Config.init(args, parser)).execute()


def run(args):
    return DatabaseExecuter(Config.init(args, parser)).run()


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    return DatabaseExecuter(Config.init(args, parser)).write()



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import re
import logging

from collections import OrderedDict

from dbmanagr.wrapper import Wrapper
from dbmanagr.logger import LogWith
from dbmanagr.config import Config
from dbmanagr.sources.source import Source
from dbmanagr.utils import prefix, prefixes, remove_prefix, replace_wildcards
from dbmanagr.utils import find_connection
from dbmanagr.queryfilter import QueryFilter
from dbmanagr.writer import Writer
from dbmanagr.jsonable import Jsonable, from_json
from dbmanagr.dto.mapper import to_dto
from dbmanagr.exception import UnknownConnectionException
from dbmanagr.exception import UnknownColumnException, UnknownTableException

from .args import parser, SqlInsertWriter

logger = logging.getLogger(__name__)


class RowItem(Jsonable):
    def __init__(self, row, include, exclude, substitutes):
        self.row = row
        self.include = include if include is not None else []
        self.exclude = exclude if exclude is not None else []
        self.substitutes = substitutes if substitutes is not None else {}

    def __hash__(self):
        return hash(self.row.row)

    def __eq__(self, o):
        return hash(self) == hash(o)

    @staticmethod
    def from_json(d):
        return RowItem(
            from_json(d['row']),
            from_json(d['include']),
            from_json(d['exclude']),
            from_json(d['substitutes'])
        )


def fk_by_a_table_name(fks):
    return dict(map(lambda (k, v): (v.a.table.name, v), fks.iteritems()))


def check_excludes(items, exclude):
    if exclude:
        for item in items:
            for x in exclude:
                matcher = re.compile(replace_wildcards(prefix(x)))
                fks = fk_by_a_table_name(item.table.foreign_keys())
                col, fk = None, None
                for k in fks.keys():
                    if matcher.match(k):
                        fk = fks[k]
                        break
                for c in item.table.columns():
                    if matcher.match(c.name):
                        col = c
                        break
                if not col and not fk:
                    raise UnknownColumnException(
                        item.table, x,
                        fks.keys() + map(
                            lambda c: c.name, item.table.columns()))
            # only check first item, as we expect all items are from the same
            # table
            break


def find_foreign_key(table, matcher):
    for _, val in table.foreign_keys().iteritems():
        if matcher.match(val.a.table.name):
            return val

    return None


def find_column(table, matcher):
    for c in table.columns():
        if matcher.match(c.name):
            return c

    return None


def add_forward_reference(includes, fk, item):
    if fk not in includes:
        includes[fk] = []
    includes[fk].append(item[fk.a.name])


def add_backward_reference(includes, fk, item):
    if fk not in includes:
        includes[fk] = []
    includes[fk].append(item[fk.b.name])


def add_column(includes, col, item):
    fk = item.table.foreign_key(col.name)
    if fk not in includes:
        includes[fk] = []
    includes[fk].append(item[fk.a.name])


def process_item(item, include, includes):
    matcher = re.compile(replace_wildcards(prefix(include)))
    col = find_column(item.table, matcher)
    fk = find_foreign_key(item.table, matcher)
    if not col and not fk:
        raise UnknownColumnException(item.table, include)
    if '.' not in include:
        # Only include column, don't include referencing rows
        #
        # Examples:
        # 1) user_id - include column user_id, but no referenced rows
        # 2) user_id. - include column user_id with referencing user
        #    row
        # 3) user_id.article - include column user_id with referenced
        #    articles
        return

    if fk:
        if fk.a.table.name == item.table.name:
            add_forward_reference(includes, fk, item)
        else:
            add_backward_reference(includes, fk, item)

    if col and col.name in item.table.foreign_keys():
        add_column(includes, col, item)


@LogWith(logger, log_args=False, log_result=False)
def create_items(connection, items, include, exclude, substitutes):
    results_pre = []
    results_post = []
    includes = {}

    check_excludes(items, exclude)

    item = None
    for item in items:
        for i in include:
            process_item(item, i, includes)

    if item is not None:
        for fk in includes.keys():
            if fk.a.table.name == item.table.name:
                # forward references, must be in pre
                results_pre += create_items(
                    connection,
                    connection.rows(
                        fk.b.table,
                        QueryFilter(fk.b.name, 'in', includes[fk]),
                        limit=-1,
                        simplify=False),
                    remove_prefix(fk.a.name, include),
                    remove_prefix(fk.a.name, exclude),
                    remove_prefix(fk.a.name, substitutes))
            else:
                # backward reference, must be in post
                results_post += create_items(
                    connection,
                    connection.rows(
                        fk.a.table,
                        QueryFilter(fk.a.name, 'in', includes[fk]),
                        limit=-1,
                        simplify=False),
                    remove_prefix(fk.a.table.name, include),
                    remove_prefix(fk.a.table.name, exclude),
                    remove_prefix(fk.a.table.name, substitutes))

    return results_pre + map(
        lambda i: RowItem(
            to_dto(i),
            prefixes(include),
            prefixes(exclude),
            prefixes(substitutes)),
        items) + results_post


class DatabaseExporter(Wrapper):
    """The main class"""
    def __init__(self, options):
        Wrapper.__init__(self, options)

        if options.formatter:
            Writer.set(options.formatter(options))
        else:
            Writer.set(SqlInsertWriter(options))

    def execute(self):
        """The main method that splits the arguments and starts the magic"""
        options = self.options

        # search exact match of connection
        connection, opts = find_connection(
            Source.connections(),
            options,
            lambda con, opts: con.matches(opts))

        if connection is None:
            raise UnknownConnectionException(
                options.uri,
                map(lambda c: c.autocomplete(), Source.connections()))

        if opts.show not in ('values', 'columns') or opts.filter is None:
            raise Exception('Specify the complete URI to a table')

        try:
            connection.connect(opts.database)
            tables = connection.tables()
            if opts.table not in tables:
                raise UnknownTableException(opts.table, tables.keys())
            table = tables[opts.table]
            items = create_items(
                connection,
                connection.rows(
                    table,
                    opts.filter,
                    opts.limit,
                    simplify=False),
                opts.include,
                opts.exclude,
                opts.substitutes)
            # remove duplicates
            return list(OrderedDict.fromkeys(items))
        finally:
            connection.close()


def execute(args):
    """
    Directly calls the execute method and avoids using the wrapper
    """
    return DatabaseExporter(Config.init(args, parser)).execute()


def run(args):
    return DatabaseExporter(Config.init(args, parser)).run()


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    return DatabaseExporter(Config.init(args, parser)).write()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import sys
import BaseHTTPServer
import json
import urllib2
import logging
import time
import traceback

from dbmanagr.jsonable import Jsonable, as_json
from dbmanagr.utils import mute_stderr

logger = logging.getLogger(__name__)


class Encoder(json.JSONEncoder):
    def default(self, obj):  # pylint: disable=method-hidden
        if isinstance(obj, Jsonable):
            return obj.as_json()
        return as_json(obj)


class DaemonHTTPServer(BaseHTTPServer.HTTPServer):
    def __init__(self, *args, **kwargs):
        BaseHTTPServer.HTTPServer.__init__(self, *args, **kwargs)
        self.active = True

    def serve_forever(self, poll_interval=0.5):
        while self.active:
            self.handle_request()


class DaemonHTTPRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
    def do_POST(self):  # noqa
        from dbmanagr.command import navigator, exporter, differ, executer
        from dbmanagr.command import grapher

        commands = {
            'navigator': navigator,
            'exporter': exporter,
            'differ': differ,
            'executer': executer,
            'grapher': grapher
        }

        parts = self.path.split('/')
        command = parts[1]
        if command == 'server-status':
            self.send_response(200)
            self.end_headers()
            return
        if command == 'server-stop':
            self.send_response(200)
            self.end_headers()
            self.server.active = False
            return
        if command not in commands:
            self.send_error(404)
            return

        args = json.loads(self.rfile.read(
            int(self.headers.getheader('content-length'))))

        try:
            items = mute_stderr(commands[command].execute)(args)
            self.send_response(200)
            self.send_header('Content-Type', 'application/json')
            self.end_headers()
            self.wfile.write(json.dumps(items, cls=Encoder))
        except BaseException as e:
            logger.debug(e)
            self.send_response(500)
            self.send_header('Content-Type', 'application/json')
            self.end_headers()
            self.wfile.write(json.dumps({
                '__cls__': str(e.__class__.__name__),
                'message': e.message,
                'traceback': as_json(traceback.extract_tb(sys.exc_info()[2]))
            }))

    def log_message(self, format_, *args):
        logger.info(format_, args)


def is_running(config):
    try:
        urllib2.urlopen(
            'http://{host}:{port}/server-status'.format(
                host=config.host,
                port=config.port),
            '')
    except BaseException:
        return False

    return True


def start_server(config):
    try:
        httpd = DaemonHTTPServer(
            (config.host, config.port), DaemonHTTPRequestHandler)
        if os.fork() == 0:
            # Child process
            httpd.serve_forever()
            sys.exit(0)
        return True
    except BaseException:
        pass
    return False


def start(config):
    sys.stdout.write('Starting server... ')

    if start_server(config):
        sys.stdout.write('OK\n')
    else:
        sys.stdout.write('already running\n')


def stop(config):
    sys.stdout.write('Stopping server... ')
    try:
        urllib2.urlopen(
            'http://{host}:{port}/server-stop'.format(
                host=config.host,
                port=config.port),
            '')
    except BaseException:
        sys.stdout.write('failed\n')
    else:
        sys.stdout.write('OK\n')


def restart(config):
    stop(config)
    time.sleep(1)
    start(config)


def status(config):
    if is_running(config):
        sys.stdout.write('Status: online\n')
    else:
        sys.stdout.write('Status: offline\n')



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

# import shelve
import logging

from dbmanagr.logger import LogWith
from dbmanagr.model.databaseconnection import UriDatabaseConnection
from dbmanagr.model.database import Database
from dbmanagr.model.table import Table
from dbmanagr.model.tablecomment import TableComment

DATABASES_QUERY = """select
        db.datname as database_name
    from
        pg_database db,
        pg_roles r
    where
        db.datistemplate = false
        and r.rolname = '%s'
        and (
            r.rolsuper
            or pg_catalog.pg_get_userbyid(db.datdba) = r.rolname
        )
    order by 1"""
TABLES_QUERY = """select
        t.table_name as tbl,
        obj_description(c.oid) as comment,
        pg_catalog.pg_get_userbyid(c.relowner) as owner,
        pg_size_pretty(pg_total_relation_size(io.relid)) as size
    from
        information_schema.tables t,
        pg_class c,
        pg_catalog.pg_statio_user_tables io
    where
        t.table_schema = 'public'
        and t.table_name = c.relname
        and io.relname = t.table_name
        and c.relkind = 'r'
    order by t.table_name"""

logger = logging.getLogger(__name__)


class PostgreSQLConnection(UriDatabaseConnection):
    def __init__(self, uri, host, port, path, user, password):
        UriDatabaseConnection.__init__(
            self,
            dbms='postgresql',
            database=path,
            uri=uri,
            host=host,
            port=port,
            user=user,
            password=password,
            subtitle='PostgreSQL Connection')
        self._databases = None

    def __repr__(self):
        return '%s@%s/%s' % (
            self.user, self.host, self.database if self.database != '*' else ''
        )

    def databases(self):
        # does not yet work with sqlalchemy...
        if self._databases is None:
            self._databases = map(
                lambda row: Database(self, row[0]),
                self.execute(DATABASES_QUERY % self.user))

        return self._databases

    @LogWith(logger)
    def init_tables(self):  # pragma: no cover
        # sqlalchemy does not yet provide reflecting comments

        self._tables = {}
        self._comments = {}

        result = self.execute(TABLES_QUERY)
        for row in result:
            self._tables[row[0]] = Table(
                self.entity(row[0]),
                self.autocomplete(),
                row[2],
                row[3])
            self._comments[row[0]] = TableComment(row[1])

        self.init_foreign_keys()



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import textwrap

from collections import OrderedDict

from dbmanagr.writer import FormatWriter
from dbmanagr.formatter import Formatter, DefaultFormatter
from dbmanagr.utils import to_yaml_type


class VerboseGraphFormatter(DefaultFormatter):
    def __init__(self, options=None):
        DefaultFormatter.__init__(self)
        self.options = options

    def format_column_node(self, node):
        return node.format_verbose(self.options.verbose)


class GraphvizFormatter(DefaultFormatter):
    def format_name_node(self, item):
        return u'  root={0};'.format(item)

    def format_table_node(self, item):
        columns = map(
            lambda (i, it): u'<{0}> {0}'.format(it.name),
            enumerate(item.table.columns))
        return u'  {0} [shape="record" label="{0}| {1}"];'.format(
            item.table.name, '| '.join(columns))

    def format_foreign_key_node(self, item):
        return u'  {fk.a.tablename}:{fk.a.name} '\
            u'-> {fk.b.tablename}:{fk.b.name} [];'.format(fk=item.fk)


class GraphWriter(FormatWriter):
    def __init__(self, options=None):
        FormatWriter.__init__(self, u'{0}\n')
        if options.verbose > 0:
            Formatter.set(VerboseGraphFormatter(options))
        else:
            Formatter.set(DefaultFormatter())


class YamlFormatter(DefaultFormatter):
    def format_name_node(self, node):
        return """{indent_}# root element
{indent_}{name}:""".format(
            indent_=(node.indent)*4*' ',
            **node.__dict__)

    def format_column_node(self, node):
        return textwrap.dedent("""
                {indent_}# {column}
                {indent_}{column.name}: # !!{type_} {nullable}""").format(
            indent_=(node.indent+1)*4*' ',
            type_=to_yaml_type(node.column.type),
            nullable={True: 'optional', False: 'mandatory'}.get(
                node.column.nullable),
            **node.__dict__)

    def format_foreign_key_node(self, node):
        if node.fk.a.tablename == node.parent.name:
            # A forward reference
            return textwrap.dedent(u"""
                    {indent_}# → {fk.a.name} → {fk.b}
                    {indent_}{fk.a.name}: # !!{type_} {nullable}""").format(
                indent_=(node.indent+1)*4*' ',
                type_=to_yaml_type(node.fk.a.type),
                nullable={True: 'optional', False: 'mandatory'}.get(
                    node.fk.a.nullable),
                **node.__dict__)

        # A back reference
        return textwrap.dedent(u"""
                {indent_}# ↑ {fk.a} → {fk.b}
                {indent_}{fk.a.tablename}:""").format(
            indent_=(node.indent+1)*4*' ',
            **node.__dict__)


class YamlWriter(FormatWriter):
    def __init__(self, options=None):
        FormatWriter.__init__(self, u'{0}\n', u'{0}', u'')
        Formatter.set(YamlFormatter())
        self.options = options


class GraphvizWriter(FormatWriter):
    def __init__(self, include_tables=False):
        FormatWriter.__init__(self, u"""digraph dbgraph {{
{0}
}}
""")
        self.include_tables = include_tables
        Formatter.set(GraphvizFormatter())

    def filter_(self, items):
        # Removes duplicate items and keeps order
        return list(OrderedDict.fromkeys(items))


class GraphTestWriter(GraphWriter):
    pass



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

from sqlalchemy.exc import ProgrammingError, DataError

from dbmanagr.logger import LogWith
from dbmanagr.querybuilder import QueryBuilder, SimplifyMapper
from dbmanagr.comment import create_comment
from dbmanagr.exception import UnknownColumnException
from dbmanagr.model.baseitem import BaseItem
from dbmanagr.model.column import create_column
from dbmanagr.model.row import Row
from dbmanagr.model import DEFAULT_LIMIT

OPTION_URI_VALUE_FORMAT = '%s%s/?%s'

logger = logging.getLogger(__name__)


class Table(BaseItem):
    def __init__(
            self,
            entity=None,
            uri=None,
            owner=None,
            size=None,
            name=None,
            primary_key=None,
            columns=None):
        self.name = None
        if entity is not None:
            self.name = entity.name
        elif name is not None:
            self.name = name
        self._entity = entity
        self.uri = uri
        self.owner = owner
        self.size = size

        if entity is not None:
            self._columns = map(
                lambda c: create_column(self, str(c.name), c), entity.columns)
        elif columns is not None:
            self._columns = map(
                lambda c: create_column(self, c), columns)
        else:
            self._columns = None
        self._fks = {}

        self.primary_key = primary_key

    def __repr__(self):
        return self.name

    def autocomplete(self):
        return self.autocomplete_()

    def autocomplete_(
            self, column=None, value=None, format_=OPTION_URI_VALUE_FORMAT):
        """Retrieves the autocomplete string for the given column and value"""

        if column is None:
            return u'%s%s?' % (self.uri, self.name)

        tablename = self.name
        if type(value) is buffer:
            value = '[BLOB]'
        else:
            value = u'%s=%s' % (column, value)

        return format_ % (self.uri, tablename, value)

    def entity(self):
        return self._entity

    def columns(self, needle=None):
        """Retrieves columns of table with optional filter applied"""

        if needle is None:
            return self._columns

        return filter(lambda c: needle in c.name, self._columns)

    def column(self, name):
        if type(name) is int:
            return self._columns[name]

        for col in self._columns:
            if col.name == name:
                return col

        return None

    @LogWith(logger, log_result=False, log_args=False)
    def rows(
            self,
            connection,
            filter_=None,
            limit=DEFAULT_LIMIT,
            simplify=None):
        """Retrieves rows from the table with the given filter applied"""

        comment = None
        order = []

        if simplify is None:
            simplify = False

        if simplify:
            comment = connection.comment(self.name)
            order = comment.order

        builder = QueryBuilder(
            connection,
            self,
            filter_=filter_,
            order=order,
            limit=limit,
            simplify=simplify)

        mapper = None
        if simplify:
            mapper = SimplifyMapper(
                self,
                comment=create_comment(
                    self,
                    comment,
                    builder.counter,
                    builder.aliases,
                    None))

        try:
            result = connection.queryall(
                builder.build(),
                mapper=mapper)
        except (DataError, ProgrammingError, UnknownColumnException,
                UnicodeEncodeError):  # pragma: no cover
            raise
        except BaseException as e:  # pragma: no cover
            logger.error(e, exc_info=1)  # pragma: no cover
            import sys  # pragma: no cover
            # pylint: disable=raising-non-exception
            raise type(e), type(e)(
                u'{} (check comment on table {})'.format(e.message, self.name)
            ), sys.exc_info()[2]  # pragma: no cover

        return map(lambda row: Row(self, row), result)

    def foreign_keys(self):
        return self._fks

    def foreign_key(self, name):
        if name in self._fks:
            return self._fks[name]
        return None

    def set_foreign_key(self, name, value):
        self._fks[name] = value

    def title(self):
        return self.name

    def subtitle(self):
        if self.owner and self.size:
            return u'Owner: %s (%s)' % (self.owner, self.size)
        return u'Table'



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#


class DefaultFormatter(object):
    def __init__(self):
        pass

    def format(self, item):
        return unicode(item)

    def format_item(self, item):
        return self.format(item)

    def format_row(self, item):
        return u', '.join(map(unicode, item))

    def format_column(self, item):
        return self.format(item)

    def format_node(self, item):
        return unicode(item)

    def format_column_node(self, item):
        return self.format(item)

    def format_name_node(self, item):
        return self.format(item)

    def format_table_node(self, item):
        return self.format(item)

    def format_foreign_key_node(self, item):
        return self.format(item)


class SimplifiedFormatter(DefaultFormatter):
    def __init__(
            self,
            default_format=u'{title}\t{subtitle}',
            item_format=u'{title}\t{subtitle}'):
        self.default_format = default_format
        self.item_format = item_format

    def format(self, item):
        return self.default_format.format(
            title=self.escape(item.title()),
            subtitle=self.escape(item.subtitle()),
            autocomplete=self.escape(item.autocomplete()),
            uid=self.escape(item.uid()),
            validity=self.escape(item.validity()),
            icon=self.escape(item.icon()),
            value=self.escape(item.value()))

    def format_row(self, item):
        return self.format(item)

    def escape(self, value):
        return value


class TestFormatter(SimplifiedFormatter):
    def format(self, item):
        return u'{title}\t{autocomplete}'.format(
            title=item.title(),
            autocomplete=item.autocomplete())


class JsonFormatter(SimplifiedFormatter):
    def __init__(self):
        SimplifiedFormatter.__init__(
            self,
            default_format=u'   {{ '
                           u'"uid": "{uid}", '
                           u'"arg": "{title}", '
                           u'"autocomplete": "{autocomplete}", '
                           u'"valid": "{validity}", '
                           u'"title": "{title}", '
                           u'"subtitle": "{subtitle}", '
                           u'"icon": "{icon}" '
                           u'}}',
            item_format=u'   {{ '
                        u'"uid": "{uid}", '
                        u'"arg": "{title}", '
                        u'"autocomplete": '
                        u'"{autocomplete}", '
                        u'"valid": "{validity}", '
                        u'"title": "{title}", '
                        u'"subtitle": "{subtitle}", '
                        u'"icon": "{icon}" '
                        u'}}')


class SimpleFormatter(SimplifiedFormatter):
    def __init__(self):
        SimplifiedFormatter.__init__(
            self,
            default_format=u"""{uid}\t{title}\t{subtitle}\t{autocomplete}""",
            item_format=u"""{uid}\t{title}\t{subtitle}\t{autocomplete}""")


class AutocompleteFormatter(SimplifiedFormatter):
    def __init__(self):
        SimplifiedFormatter.__init__(
            self,
            default_format=u"""{autocomplete}""",
            item_format=u"""{autocomplete}""")


class Formatter(object):
    formatter = DefaultFormatter()

    @staticmethod
    def set(arg):
        Formatter.formatter = arg

    @staticmethod
    def format(item):
        return Formatter.formatter.format(item)

    @staticmethod
    def format_row(item):
        return Formatter.formatter.format_row(item)

    @staticmethod
    def format_column(item):
        return Formatter.formatter.format_column(item)

    @staticmethod
    def format_node(item):
        return Formatter.formatter.format_node(item)

    @staticmethod
    def format_column_node(item):
        return Formatter.formatter.format_column_node(item)

    @staticmethod
    def format_table_node(item):
        return Formatter.formatter.format_table_node(item)

    @staticmethod
    def format_name_node(item):
        return Formatter.formatter.format_name_node(item)

    @staticmethod
    def format_foreign_key_node(item):
        return Formatter.formatter.format_foreign_key_node(item)



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import logging
import yaml

from dbmanagr.wrapper import Wrapper
from dbmanagr.config import Config
from dbmanagr.logger import LogWith
from dbmanagr.writer import Writer
from dbmanagr.utils import is_node, to_ref, to_forward_ref, shell_escape

from .args import parser
from .writer import ArgumentWriter

logger = logging.getLogger(__name__)


@LogWith(logger)
def consume(tree, parent, includes, excludes, substitutes):
    if is_node(tree, 'includes'):
        dfs(tree['includes'], parent, includes,
            lambda ref, v: to_forward_ref(ref))
    if is_node(tree, 'excludes'):
        dfs(tree['excludes'], parent, excludes,
            lambda ref, v: ref)
    if is_node(tree, 'substitutes'):
        dfs(tree['substitutes'], parent, substitutes,
            lambda ref, v: u'{0}={1}'.format(ref, shell_escape(v)))


@LogWith(logger)
def dfs(tree, parent, selection, conversion):
    """Depth first search for selection"""

    for (k, v) in tree.iteritems():
        ref = to_ref(parent, k)
        if type(v) is dict:
            dfs(v, ref, selection, conversion)
        else:
            selection.append(conversion(ref, v))


class DatabaseArgumentor(Wrapper):
    """The main class"""
    def __init__(self, options):
        Wrapper.__init__(self, options)

        if options.formatter:
            Writer.set(options.formatter(options))
        else:
            Writer.set(ArgumentWriter(options))

    @LogWith(logger)
    def execute(self):
        """The main method that splits the arguments and starts the magic"""
        options = self.options

        config = yaml.safe_load(options.infile)

        logger.debug('Config file: %s', config)

        includes = []
        excludes = []
        substitutes = []

        consume(config, None, includes, excludes, substitutes)

        return (includes, excludes, substitutes)


def execute(args):
    """
    Directly calls the execute method and avoids using the wrapper
    """
    return DatabaseArgumentor(Config.init(args, parser)).execute()


def run(args):
    return DatabaseArgumentor(Config.init(args, parser)).run()


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    return DatabaseArgumentor(Config.init(args, parser)).write()



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

from sqlalchemy import create_engine, MetaData
from sqlalchemy.exc import OperationalError
from sqlalchemy.engine import reflection

from dbmanagr.logger import LogWith, LogTimer
from dbmanagr.model import DEFAULT_LIMIT
from dbmanagr.model.column import create_column
from dbmanagr.model.baseitem import BaseItem
from dbmanagr.model.foreignkey import ForeignKey
from dbmanagr.model.database import Database
from dbmanagr.model.table import Table
from dbmanagr.model.tablecomment import TableComment
from dbmanagr.utils import escape_statement

logger = logging.getLogger(__name__)


class DatabaseConnection(BaseItem):
    def __init__(self, **kwargs):
        self.dbms = kwargs.get('dbms', None)
        self.database = kwargs.get('database', None)
        self.uri = kwargs.get('uri', None)
        self._subtitle = kwargs.get('subtitle', 'Generic Connection')
        self._tables = kwargs.get('tables', None)
        self._comments = kwargs.get('comments', None)
        self._engine = None
        self._inspector = None
        self._meta = None
        self._con = None

    def title(self):
        return self.__repr__()

    def subtitle(self):
        return self._subtitle

    def autocomplete(self):
        return self.__repr__()

    def icon(self):
        return 'images/connection.png'

    def matches(self, options):  # pragma: no cover
        pass

    def connect(self, database):  # pragma: no cover
        """Is to be implemented by subclasses"""

        pass

    @LogWith(logger)
    def connect_to(self, source):
        self._engine = create_engine(source)
        self._con = self._engine.connect()

    def engine(self):
        return self._engine

    def meta(self):
        if self._meta is None:
            self._meta = MetaData()
            self._meta.reflect(bind=self._engine)

        return self._meta

    def inspector(self):
        if self._inspector is None:
            self._inspector = reflection.Inspector.from_engine(self._engine)

        return self._inspector

    def connected(self):
        return self._con is not None

    def close(self):
        if self._con:
            self._con.close()
            self._con = None

    def cursor(self):
        return self._con

    def begin(self):
        return self._con.begin()

    @LogWith(logger, log_args=False, log_result=False)
    def execute(self, query):
        cur = self.cursor()
        if not cur:
            raise Exception('Database is not connected')

        timer = LogTimer(logger, 'Execution', 'Executing:\n%s', query)
        try:
            return cur.execute(escape_statement(query))
        finally:
            timer.stop()

    @LogWith(logger, log_args=False, log_result=False)
    def queryall(self, query, mapper=None):
        timer = LogTimer(logger, 'Query all', 'Querying all:\n%s', query)
        result = query.all()
        timer.stop()

        if mapper:
            for row in result:
                mapper.map(row)

        return result

    @LogWith(logger, log_args=False, log_result=False)
    def queryone(self, query, mapper=None):
        timer = LogTimer(logger, 'Query one', 'Querying one:\n%s', query)
        result = query.one()
        timer.stop()

        if mapper:
            mapper.map(result)

        return result

    def rows(self, table, filter_=None, limit=DEFAULT_LIMIT, simplify=None):
        return table.rows(self, filter_, limit, simplify)

    def filter_(self, options):
        pass

    def databases(self):
        return map(
            lambda name: Database(self, name),
            self.inspector().get_schema_names())

    @LogWith(logger)
    def init_tables(self):
        self._tables = dict(map(
            lambda table: (table, Table(
                self.entity(table), self.autocomplete())),
            self.meta().tables))
        logger.debug('Tables: %s', self._tables)
        self.init_foreign_keys()

    @LogWith(logger)
    def tables(self):
        if not self._tables:
            self.init_tables()

        return self._tables

    def table(self, tablename):
        return self.tables().get(tablename)

    def entity(self, tablename):
        return self.meta().tables[tablename]

    def init_comments(self):
        self._comments = dict(map(
            lambda k: (k, TableComment('')),
            self.tables().keys()))
        comment = self.table('_comment')
        if comment:
            # Table _comments exists, query it
            for row in self.rows(comment, limit=-1, simplify=False):
                self._comments[row['table']] = TableComment(row['comment'])

    def comments(self):
        if not self._comments:
            self.init_comments()

        return self._comments

    def comment(self, tablename):
        return self.comments().get(tablename, None)

    def init_foreign_keys(self):
        for _, t in self.meta().tables.iteritems():
            for _fk in t.foreign_keys:
                a = create_column(
                    self._tables[_fk.parent.table.name],
                    str(_fk.parent.key),
                    _fk.parent)
                b = create_column(
                    self._tables[_fk.column.table.name],
                    str(_fk.column.key),
                    _fk.column)
                fk = ForeignKey(a, b)
                self._tables[a.table.name].set_foreign_key(a.name, fk)
                self._tables[b.table.name].set_foreign_key(str(a), fk)

    def __repr__(self):  # pragma: no cover
        raise Exception('Should be overridden in subclass')

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):  # pragma: no cover
        return self.__repr__() == other.__repr__()

    def __hash__(self):
        return hash(self.__repr__())


class UriDatabaseConnection(DatabaseConnection):
    def __init__(self, **kwargs):
        DatabaseConnection.__init__(self, **kwargs)
        self.user = kwargs.get('user', None)
        self.password = kwargs.get('password', None)
        self.host = kwargs.get('host', None)
        self.database = kwargs.get('database', None)

    @LogWith(logger)
    def connect(self, database):
        if database:
            try:
                self.connect_to(
                    self.uri.format(
                        user=self.user,
                        password=self.password,
                        host=self.host,
                        database=database))
                self.database = database
            except OperationalError:
                self.connect_to(
                    self.uri.format(
                        user=self.user,
                        password=self.password,
                        host=self.host,
                        database=''))
                database = None
        else:
            self.connect_to(
                self.uri.format(
                    user=self.user,
                    password=self.password,
                    host=self.host,
                    database=''))

    def title(self):
        return self.autocomplete()

    def autocomplete(self):
        """Retrieves the autocomplete string"""

        if self.database and self.database != '*':
            return '%s@%s/%s/' % (self.user, self.host, self.database)

        return '%s@%s/' % (self.user, self.host)

    def matches(self, options):
        if options.gen:
            return options.gen.startswith("%s@%s" % (self.user, self.host))
        return False

    def filter_(self, options):
        options = options.get(self.dbms)
        matches = True

        if options.user:
            filter_ = options.user
            if options.host is not None:
                matches = filter_ in self.user
            else:
                matches = filter_ in self.user or filter_ in self.host
        if options.host is not None:
            matches = matches and options.host in self.host

        return matches



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import logging
import time
import functools
import inspect

logger = logging.getLogger(__name__)

ENTRY_MESSAGE = u'⇢ %s({})'
EXIT_MESSAGE = u'⇠ %s [%0.3fms] = %s'


def encode(v):
    if v is None:
        return None
    if type(v) is unicode:
        return repr(v)
    if type(v) is str:
        return encode(unicode(v, 'UTF-8'))
    if type(v) is list:
        return map(encode, v)
    return encode(unicode(v))


def argtostring(k, v):
    if k == 'self':
        return k
    return '{0}={1}'.format(k, encode(v))


def log_error(e):
    sys.stderr.write('{0}: {1}\n'.format(
        sys.argv[0].split('/')[-1], e))


class LogWith(object):
    '''Logging decorator that allows you to log with a specific logger.
'''

    def __init__(self, logger, log_args=True, log_result=True):
        self.logger = logger
        self.log_args = log_args
        self.log_result = log_result

    def __call__(self, f):
        '''Returns a wrapper that wraps function f. The wrapper will log the
entry and exit points of the function with logging.DEBUG level.
'''

        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            if self.logger.getEffectiveLevel() <= logging.DEBUG:
                if self.log_args:
                    # Keeps order of args intact
                    cargs = inspect.getcallargs(f, *args)
                    fargs = map(
                        lambda k: argtostring(k, cargs[k]),
                        inspect.getargspec(f).args)
                    # Adds keyword arguments
                    fargs += map(lambda (k, v): encode(v), kwargs)

                    # Creates format for the log message
                    formats = map(lambda arg: '%s', fargs)
                    formats += map(lambda (k, v): '{}=%s'.format(k), kwargs)

                    # Do the logging
                    self.logger.debug(
                        ENTRY_MESSAGE.format(', '.join(formats)),
                        f.__name__, *fargs)
                else:
                    # Do the logging
                    self.logger.debug(
                        ENTRY_MESSAGE.format('<omitted>'),
                        f.__name__)

                # Starts the stopwatch
                start = time.time()
                # Invokes the method/function
                result = f(*args, **kwargs)

                if self.log_result:
                    # Logs returned value and duration of the call
                    self.logger.debug(
                        EXIT_MESSAGE,
                        f.__name__,
                        (time.time() - start) * 1000.0,
                        encode(result))
                else:
                    self.logger.debug(
                        EXIT_MESSAGE,
                        f.__name__,
                        (time.time() - start) * 1000.0,
                        '<omitted>')

                return result
            else:
                return f(*args, **kwargs)
        return wrapper


class LogTimer(object):
    def __init__(self, logger, subject, prolog=None, *pargs):
        self.logger = logger
        self.subject = subject
        self.start = time.time()
        if prolog is not None:
            self.logger.info(prolog, *pargs)

    def stop(self):
        self.logger.info(
            '%s took: %0.6fs',
            self.subject,
            time.time() - self.start)



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

import xml.etree.ElementTree as ET
from urlparse import urlparse
from os.path import isfile

from dbmanagr.sources.source import Source

logger = logging.getLogger(__name__)


class DBExplorerSource(Source):
    def __init__(self, driver, file_, scheme, con_creator):
        Source.__init__(self)
        self.driver = driver
        self.file = file_
        self.scheme = scheme
        self.con_creator = con_creator

    def list(self):
        if not isfile(self.file):
            from os.path import realpath
            logger.warn(
                'File %r does not exist (this file: %r)',
                self.file, realpath(__file__))

            return self._connections

        if not self._connections:
            try:
                tree = ET.parse(self.file)
            except Exception as e:
                logger.warn(
                    'Error parsing dbExplorer config file: %s', e.message)
                return []
            root = tree.getroot()
            for c in root.iter('connection'):
                url = urlparse(c.find('url').text.replace('jdbc:', ''))
                if url.scheme == self.scheme:
                    host = url.netloc.split(':')[0]
                    port = 3306
                    database = '*'
                    usernode = c.find('user')
                    if usernode is not None:
                        user = usernode.text
                    else:
                        user = None
                    passwordnode = c.find('password')
                    if passwordnode is not None:
                        password = passwordnode.text
                    else:
                        password = None
                    connection = self.con_creator(
                        self.driver, host, port, database, user, password)
                    self._connections.append(connection)

        return self._connections



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import math
import datetime

from urlparse import urlparse
from sqlalchemy import Boolean, Float, Integer
from decimal import Decimal

from dbmanagr.logger import LogWith
from dbmanagr.queryfilter import QueryFilter, OrOp, AndOp
from dbmanagr.utils import escape_statement

OR_OPERATOR = '|'
AND_OPERATOR = '&'
OPERATORS = ['>=', '<=', '!=', '=', '~', '*', '>', '<', ':']

OPTION_URI_FORMAT = '%s@%s/%s'

logger = logging.getLogger(__name__)


def escape_keyword(keyword):
    if keyword in ['user', 'table', 'column']:
        return '"%s"' % keyword
    return keyword


@LogWith(logger)
def restriction(alias, column, operator, value, map_null_operator=True):
    if not column:
        raise Exception('Parameter column may not be None!')
    if operator in ['=', '!='] and (value is None or value == 'null'):
        if map_null_operator:
            operator = {
                '=': 'is',
                '!=': 'is not'
            }.get(operator)
        value = None
    if column.tablename and alias is not None:
        return u"{0}.{1} {2} {3}".format(
            alias,
            escape_keyword(column.name),
            operator,
            format_value(column, value))
    return u'{0} {1} {2}'.format(
        escape_keyword(column.name),
        operator,
        format_value(column, value))


def format_value(column, value):
    if value is None or (type(value) is float and math.isnan(value)):
        return 'null'
    if type(value) is list:
        return '({0})'.format(
            ', '.join([format_value(column, v) for v in value]))
    if type(value) in [datetime.datetime, datetime.date, datetime.time]:
        return "'%s'" % value
    if type(value) is buffer:
        return u"'[BLOB]'"
    if column is None:
        try:
            return '%d' % int(value)
        except ValueError:
            return u"'%s'" % value
    if (isinstance(column.type, Boolean)
            and (type(value) is bool or value in ['true', 'false'])
            or type(value) is bool):
        return '%s' % str(value).lower()
    if isinstance(column.type, Float) or type(value) is float:
        try:
            return '%f' % float(value)
        except ValueError:
            pass
    if isinstance(column.type, Integer) or type(value) in (int, Decimal):
        try:
            return '%d' % int(value)
        except ValueError:
            pass
    return u"'%s'" % escape_statement(value).replace("'", "''")


def parse_filter(s):
    _or = OrOp()
    for t in s.split(OR_OPERATOR):
        _and = AndOp()
        for term in t.split(AND_OPERATOR):
            found = False
            for operator in OPERATORS:
                if operator in term:
                    f = term.split(operator, 1)
                    lhs = f[0]
                    rhs = f[1] if len(f) > 1 else None
                    if operator == ':':
                        rhs = rhs.split(',')
                    _and.append(QueryFilter(lhs, operator, rhs))
                    found = True
                    break
            if not found:
                _and.append(QueryFilter(term))
        _or.append(_and)
    return _or


class Options(object):
    parser = {}

    def __init__(self, argv, parser):
        logger.info('Called with params: %s', argv)

        self.opts = {}
        self.argv = argv
        self.uri = None
        self.logfile = None
        self.loglevel = None
        self.database = None
        self.table = None
        self.column = ''
        self.operator = None
        self.filter = None
        self.show = 'connections'
        self.simplify = False
        self.prog = parser.prog
        self.daemon = False
        self.command = None
        self.left = None
        self.right = None

        args = parser.parse_args(argv)

        if hasattr(args, 'include'):
            args.include = args.include.split(',') if args.include else []
        if hasattr(args, 'exclude'):
            args.exclude = args.exclude.split(',') if args.exclude else []

        self.__dict__.update(args.__dict__)

        self.update_parsers()

    def update_parsers(self):
        for k in Options.parser:
            self.opts[k] = Options.parser[k].parse(self)

    def get(self, parser):
        return self.opts[parser]

    def escape_keyword(self, keyword):
        return escape_keyword(keyword)

    def restriction(
            self, alias, column, operator, value, map_null_operator=True):
        return restriction(alias, column, operator, value, map_null_operator)

    def format_value(self, column, value):
        return format_value(column, value)

    def __setattr__(self, name, value):
        self.__dict__[name] = value
        if self.opts:
            for k in self.opts.keys():
                self.opts[k].__dict__[name] = value

    def __repr__(self):
        return self.__dict__.__repr__()


class OptionsParser(object):
    def create_driver(self):  # pragma: no cover
        pass

    def parse(self, source):  # pragma: no cover
        pass


class FileOptionsParser(OptionsParser):
    def parse(self, source):
        driver = self.create_driver()
        driver.__dict__.update(source.__dict__)
        if driver.uri:
            uri = driver.uri
            url = urlparse('url://%s' % uri)
            paths = url.path.split('?')[0].split('/')

            if len(paths) > 1:
                driver.table = paths[1]
            if '?' in uri:
                driver.filter = parse_filter(url.query)
                paths.append(url.query)

            driver.show_code = len(paths)
            driver.show = {
                1: 'connections',
                2: 'tables',
                3: 'columns',
                4: 'values'
            }.get(driver.show_code, 'connections')

        return driver


class UriOptionsParser(OptionsParser):
    def parse(self, source):
        driver = self.create_driver()
        driver.__dict__.update(source.__dict__)
        if driver.uri:
            uri = driver.uri
            if '@' not in uri:
                uri += '@'
            url = urlparse('url://%s' % uri)
            locs = url.netloc.split('@')
            paths = url.path.split('?')[0].split('/')

            if len(locs) > 0:
                driver.user = locs[0]
            if len(locs) > 1 and '@' in driver.uri:
                driver.host = locs[1]
            if len(paths) > 1:
                driver.database = paths[1]
            if len(paths) > 2:
                driver.table = paths[2]
            if '?' in uri:
                driver.filter = parse_filter(url.query)
                paths.append(url.query)

            driver.show_code = len(paths)
            driver.show = {
                1: 'connections',
                2: 'databases',
                3: 'tables',
                4: 'columns',
                5: 'values'
            }.get(driver.show_code, 'connections')

        if driver.user and driver.host:
            driver.gen = OPTION_URI_FORMAT % (
                driver.user,
                driver.host,
                driver.table if driver.table else ''
            )

        return driver



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from .baseitem import BaseItem

OPTION_URI_DATABASE_FORMAT = '{}/'
AUTOCOMPLETE_FORMAT = '{connection.user}@{connection.host}/{database}'


class Database(BaseItem):
    """The database used with the given connection"""

    def __init__(
            self,
            connection,
            name,
            autocomplete_format=AUTOCOMPLETE_FORMAT):
        self._connection = connection
        self.name = name
        self._autocomplete_format = autocomplete_format
        self.uri = repr(self)

    def __repr__(self):
        return self._autocomplete_format.format(
            connection=self._connection,
            database=self.name)

    def autocomplete(self):
        """Retrieves the autocomplete string"""

        return OPTION_URI_DATABASE_FORMAT.format(self.uri)



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import logging

from collections import deque

from dbmanagr.wrapper import Wrapper
from dbmanagr.config import Config
from dbmanagr.sources.source import Source
from dbmanagr.logger import LogWith
from dbmanagr.utils import find_connection, to_dict, selection, is_included, \
    is_excluded
from dbmanagr.writer import Writer
from dbmanagr.dto.mapper import to_dto
from dbmanagr.exception import UnknownTableException, \
    UnknownConnectionException

from .args import parser
from .writer import GraphWriter, GraphvizWriter
from .node import ColumnNode, ForeignKeyNode, NameNode, TableNode

logger = logging.getLogger(__name__)


def include_forward_references(
        table, head, consumed, include, exclude, indent, opts):
    found = False
    for col in filter(
            lambda col: (
                not is_excluded(col.name, exclude) or
                is_included(col.name, include)),
            table.columns()):
        logger.debug('col.name %s not in exclude %s', col.name, exclude)
        fk = table.foreign_key(col.name)
        # logger.debug('consumed=%s', consumed)
        if not fk:
            if opts.include_columns:
                # Add column
                head.append(ColumnNode(col, indent))
        elif fk.a.table.name == table.name:
            # Collects the forward references
            logger.debug(
                'adds forward reference: fk=%s, include=%s',
                fk, include)
            head.append(ForeignKeyNode(fk, table, indent))
            if (is_included(fk.a.name, include)
                    or (opts.recursive and fk.b.table.name not in consumed)):
                # logger.debug('adds table=%s', fk.b.table)
                head.append(TableNode(
                    fk.b.table,
                    include=selection(fk.a.name, include),
                    exclude=selection(fk.a.name, exclude),
                    indent=indent + 1))
                found = True

    return found


def include_back_references(
        table, head, consumed, include, exclude, indent, opts):
    found = False
    for _, fk in filter(
            lambda (key, fk): (
                fk.b.table.name == table.name
                and (
                    not is_excluded(fk.a.table.name, exclude)
                    or is_included(fk.a.table.name, include))),
            table.foreign_keys().iteritems()):
        logger.debug(
            'adds back reference: fk=%s, include=%s',
            fk, include)
        head.append(ForeignKeyNode(fk, table, indent))
        if (is_included(fk.a.table.name, include)
                or (opts.recursive and fk.a.table.name not in consumed)):
            # Collects the back references
            # logger.debug('adds table=%s', fk.a.table)
            head.append(TableNode(
                fk.a.table,
                include=selection(fk.a.table.name, include),
                exclude=selection(fk.a.table.name, exclude),
                indent=indent + 1))
            found = True

    return found


def bfs(start, include, exclude, indent, opts):
    logger.debug(
        'bfs(start=%s, include=%s, exclude=%s, indent=%d)',
        start, include, exclude, indent)

    head = [TableNode(start, include, exclude, indent)]
    consumed = []
    found = True
    while found:
        if opts.max_depth > 0 and indent > opts.max_depth:
            break

        found = False
        tail = deque(head)
        head = []

        while tail:
            node = tail.popleft()

            if isinstance(node, TableNode) and node.indent >= indent:
                # consume node
                table = node.table
                if opts.recursive and table.name in consumed:
                    continue  # pragma: no cover
                if opts.formatter is GraphvizWriter:
                    # Keep node as we need to display its columns in the graph
                    head.append(node)

                include = node.include
                exclude = node.exclude

                consumed.append(table.name)

                logger.debug(
                    'consume table=%s, include=%s, exclude=%s, consumed=%s, '
                    'indent=%d',
                    table, include, exclude, consumed, indent)

                if include_forward_references(
                        table, head, consumed, include, exclude, indent, opts):
                    found = True

                logger.debug(
                    'opts.include_back_references: %s',
                    opts.include_back_references)
                if opts.include_back_references and include_back_references(
                        table, head, consumed, include, exclude, indent, opts):
                    found = True
            else:
                head.append(node)
        indent += 1

    return head


class DatabaseGrapher(Wrapper):
    """The main class"""
    def __init__(self, options):
        Wrapper.__init__(self, options)

        if options.formatter:
            Writer.set(options.formatter(options))
        else:
            Writer.set(GraphWriter(options))

    @LogWith(logger)
    def execute(self):
        """The main method that splits the arguments and starts the magic"""
        options = self.options

        # search exact match of connection
        connection, opts = find_connection(
            Source.connections(),
            options,
            lambda con, opts: con.matches(opts))

        if connection is None:
            raise UnknownConnectionException(
                options.uri,
                map(lambda c: c.autocomplete(), Source.connections()))

        if opts.show not in ['tables', 'columns', 'values']:
            raise Exception('Specify the complete URI to a table')

        return to_dto(self.build(connection, opts))

    @LogWith(logger)
    def build(self, connection, opts):
        try:
            connection.connect(opts.database)
            tables = connection.tables()
            if opts.table not in tables:
                raise UnknownTableException(opts.table, tables.keys())
            table = tables[opts.table]
            nodes = []
            indent = 0
            if opts.include_driver:
                nodes.append(
                    NameNode(connection.dbms, indent=indent))
                indent += 1
            if opts.include_connection:
                nodes.append(NameNode(str(connection), indent=indent))
                indent += 1
            if opts.include_database and opts.database:
                nodes.append(NameNode(opts.database, indent=indent))
                indent += 1
            nodes.append(NameNode(table.name, indent=indent))
            nodes += bfs(
                table,
                include=to_dict(opts.include),
                exclude=to_dict(opts.exclude),
                indent=indent,
                opts=opts)

            def include_node(item):
                if opts.formatter is GraphvizWriter:
                    if isinstance(item, TableNode):
                        return True
                    if isinstance(item, ColumnNode):
                        return False
                elif opts.include_columns and isinstance(
                        item, ColumnNode):
                    return True
                return not isinstance(item, TableNode)

            return filter(include_node, nodes)
        finally:
            connection.close()


def execute(args):
    """
    Directly calls the execute method and avoids using the wrapper
    """
    return DatabaseGrapher(Config.init(args, parser)).execute()


def run(args):
    return DatabaseGrapher(Config.init(args, parser)).run()


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    return DatabaseGrapher(Config.init(args, parser)).write()



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from dbmanagr.args import parent_parser, format_group, create_parser
from dbmanagr.args import CommaSeparatedPlainList, CommaSeparatedStringList

from .writer import StatementActivityWriter

parent = parent_parser()

group = format_group(parent)
group.add_argument(
    '-D',
    '--default',
    help='output format: default',
    dest='formatter',
    action='store_const',
    const=StatementActivityWriter)

parser = create_parser(
    prog='dbstac',
    description='Lists the statements and transactions currently active in '
                'the DBMS.',
    parents=[parent])
parser.add_argument(
    'uri',
    help='the URI to a DBMS')
parser.add_argument(
    'pattern',
    default=None,
    help='shows only queries that match the pattern (default: no pattern)',
    nargs='?')
parser.add_argument(
    '-u',
    '--username',
    default='',
    help='filter the user')
parser.add_argument(
    '-p',
    '--pids',
    action=CommaSeparatedPlainList,
    default=-1,
    help='filter the process ids of the activity; multiple columns can be '
         'specified by separating them with a comma')
parser.add_argument(
    '-s',
    '--states',
    action=CommaSeparatedStringList,
    default="''",
    help='filter the states of the activity; multiple columns can be '
         'specified by separating them with a comma')
parser.add_argument(
    '-a',
    '--all',
    dest='disabled',
    default=False,
    action='store_true',
    help='show all activity, even those with insufficient privileges')
parser.add_argument(
    '-v',
    '--verbose',
    action='count',
    help='specify the verbosity of the output, increase the number of '
         'occurences of this option to increase verbosity')



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

STAT_ACTIVITY = """select
        sa.datname database_name,
        sa.pid pid,
        sa.usename username,
        coalesce(sa.client_addr || ':' || sa.client_port, '') as client,
        sa.xact_start transaction_start,
        sa.query_start query_start,
        sa.state state,
        regexp_replace(sa.query, '\s+', ' ', 'g') query,
        sa.waiting blocked,
        case
            when sa.waiting then string_agg(l.virtualxid, ',')
            else ''
        end blocked_by,
        now() - sa.xact_start as transaction_duration,
        now() - sa.query_start as query_duration
    from pg_stat_activity sa
        left join pg_locks l on l.pid = sa.pid
    where
        '{database}' in (sa.datname, '')
        and (
            sa.state in ({states})
            or array[{states}] = array['']
        )
        and (sa.state != 'disabled' or {disabled})
        and '{username}' in (sa.usename, '')
        and (
            sa.pid in ({pids})
            or array[{pids}] = array[-1]
        )
        and (
            '{pattern}' = ''
            or sa.query like '%{pattern}%'
        )
    group by 1, 2, 3, 4, 5, 6, 7, 8, 9
    order by sa.datname
"""



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from dbmanagr.model.databaseconnection import UriDatabaseConnection

AUTOCOMPLETE_FORMAT = '{user}@{host}/{database}'


class MySQLConnection(UriDatabaseConnection):
    def __init__(self, uri, host, port, path, user, password):
        UriDatabaseConnection.__init__(
            self,
            dbms='mysql',
            database=path,
            uri=uri,
            host=host,
            port=port,
            user=user,
            password=password,
            subtitle='MySQL Connection')

    def __repr__(self):
        return AUTOCOMPLETE_FORMAT.format(
            user=self.user,
            host=self.host,
            database=self.database if self.database != '*' else ''
        )



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from dbmanagr.args import parent_parser, format_group, create_parser
from dbmanagr.writer import TestWriter

from .writer import SimplifiedWriter, SimpleWriter, JsonWriter
from .writer import AutocompleteWriter

parent = parent_parser(daemonable=True)

group = format_group(parent, TestWriter)
group.add_argument(
    '-D',
    '--default',
    help='output format: default',
    dest='formatter',
    action='store_const',
    const=SimplifiedWriter)
group.add_argument(
    '-S',
    '--simple',
    help='output format: simple',
    dest='formatter',
    action='store_const',
    const=SimpleWriter)
group.add_argument(
    '-J',
    '--json',
    help='output format: JSON',
    dest='formatter',
    action='store_const',
    const=JsonWriter)
group.add_argument(
    '-A',
    '--autocomplete',
    help='output format: autocomplete',
    dest='formatter',
    action='store_const',
    const=AutocompleteWriter)
parser = create_parser(
    prog='dbmanagr',
    description='A database navigation tool that shows database structure and'
                ' content',
    parents=[parent])
parser.add_argument(
    'uri',
    help='the URI to parse (format for PostgreSQL/MySQL: '
         'user@host/database/table?filter; for SQLite: '
         'databasefile.db/table?filter)',
    nargs='?')
parser.add_argument(
    '-s',
    '--simplify',
    dest='simplify',
    default=True,
    help='simplify the output',
    action='store_true')
parser.add_argument(
    '-N',
    '--no-simplify',
    dest='simplify',
    help='don\'t simplify the output',
    action='store_false')
parser.add_argument(
    '-m',
    '--limit',
    type=int,
    default=50,
    help='limit the results of the main query to this amount of rows '
         '(default: %(default)s)')



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import argparse

from dbmanagr.args import parent_parser, format_group, create_parser

from .writer import ArgumentWriter, ArgumentTestWriter, ArgumentVerboseWriter

parent = parent_parser(daemonable=True)

group = format_group(
    parent,
    ArgumentTestWriter)
group.add_argument(
    '-D',
    '--default',
    default=True,
    help='output format: human readable hierarchical text',
    dest='formatter',
    action='store_const',
    const=ArgumentWriter)
group.add_argument(
    '-V',
    help='output format: better readable text',
    dest='formatter',
    action='store_const',
    const=ArgumentVerboseWriter)

parser = create_parser(
    prog='dbargs',
    description='A tool to convert YAML config files to db{graph,export,nav} '
                'command line arguments',
    parents=[parent])
parser.add_argument(
    'infile',
    default='-',
    help='the path to the file containing the YAML config to convert',
    type=argparse.FileType('r'),
    nargs='?')
parser.add_argument(
    '-i',
    '--includes',
    default=True,
    dest='includes',
    action='store_true',
    help='show includes in output')
parser.add_argument(
    '-I',
    '--no-includes',
    dest='includes',
    action='store_false',
    help='hide includes in output')
parser.add_argument(
    '-x',
    '--excludes',
    default=True,
    dest='excludes',
    action='store_true',
    help='show excludes in output')
parser.add_argument(
    '-X',
    '--no-excludes',
    dest='excludes',
    action='store_false',
    help='hide excludes in output')
parser.add_argument(
    '-s',
    '--substitutes',
    default=True,
    dest='substitutes',
    action='store_true',
    help='show substitutes in output')
parser.add_argument(
    '-S',
    '--no-substitutes',
    dest='substitutes',
    action='store_false',
    help='hide substitutes in output')

parser.add_argument(
    '-v',
    '--verbose',
    action='count',
    help='specify the verbosity of the output, increase the number of '
         'occurences of this option to increase verbosity')



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from dbmanagr.dto import Dto
from dbmanagr.jsonable import from_json
from dbmanagr.formatter import Formatter

PRIMARY_KEY_OPTIONS = {
    True: '*',
    False: ''
}
NULLABLE_OPTIONS = {
    True: '?',
    False: '',
    None: ''
}


class BaseNode(Dto):
    def __eq__(self, o):
        return hash(self) == hash(o)

    def __hash__(self):
        return hash(str(self))

    def format(self):
        return Formatter.format_node(self)

    def format_verbose(self, verbosity=0):
        if verbosity > -1:
            return self.format()
        return None


class ColumnNode(BaseNode):
    def __init__(self, column, indent=0):
        BaseNode.__init__(self)

        self.column = column
        self.indent = indent

    def __hash__(self):
        return hash(str(self.column))

    def __str__(self):
        indent = '  ' * self.indent
        return '{0}- {1}{2}{3}'.format(
            indent,
            self.column.name,
            PRIMARY_KEY_OPTIONS.get(self.column.primary_key),
            NULLABLE_OPTIONS.get(self.column.nullable))

    def format(self):
        return Formatter.format_column_node(self)

    def format_verbose(self, verbosity=0):
        indent = '  ' * self.indent
        return '{0}- {1}'.format(indent, self.column.ddl())

    @staticmethod
    def from_json(d):
        return ColumnNode(from_json(d['column']), from_json(d['indent']))


class ForeignKeyNode(BaseNode):
    def __init__(self, fk, parent=None, indent=0):
        BaseNode.__init__(self)

        self.fk = fk
        self.parent = parent
        self.indent = indent

    def __str__(self):
        indent = '  ' * self.indent
        if self.fk.a.tablename == self.parent.name:
            return u'{0}→ {1}{3} → {2}'.format(
                indent,
                self.fk.a.name,
                self.fk.b,
                NULLABLE_OPTIONS.get(self.fk.a.nullable))
        return u'{0}↑ {1} ({2} → {3}.{4})'.format(
            indent,
            self.fk.a.tablename,
            self.fk.a.name,
            self.fk.b.tablename,
            self.fk.b.name)

    def __hash__(self):
        return hash(str(self.fk))

    def format(self):
        return Formatter.format_foreign_key_node(self)

    @staticmethod
    def from_json(d):
        return ForeignKeyNode(
            from_json(d['fk']),
            from_json(d['parent']),
            from_json(d['indent'])
        )


class TableNode(BaseNode):
    def __init__(self, table, indent=0):
        BaseNode.__init__(self)

        self.table = table
        self.indent = indent

    def __str__(self):
        return self.table.name

    def __hash__(self):
        return hash(str(self.table))

    def format(self):
        return Formatter.format_table_node(self)

    @staticmethod
    def from_json(d):
        return TableNode(
            from_json(d['table']),
            from_json(d['indent'])
        )


class NameNode(BaseNode):
    def __init__(self, name, indent=0):
        BaseNode.__init__(self)

        self.name = name
        self.indent = indent

    def __str__(self):
        return '{0}{1}'.format('  ' * self.indent, self.name)

    def __hash__(self):
        return hash(self.name)

    def format(self):
        return Formatter.format_name_node(self)

    @staticmethod
    def from_json(d):
        return NameNode(
            from_json(d['name']),
            from_json(d['indent'])
        )



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from dbmanagr.model.baseitem import BaseItem
from dbmanagr.formatter import Formatter
from dbmanagr.utils import dictminus


class Column(BaseItem):
    """A table column"""

    def __init__(self, table, name, **kwargs):
        self.table = table
        self.name = name
        self.type = None
        self.nullable = None
        self.default = None
        self.__dict__.update(kwargs)
        self.tablename = table.name
        self.uri = self.autocomplete()

    def __repr__(self):
        return '%s.%s' % (self.table.name, self.name)

    def __str__(self):
        return self.__repr__()

    def ddl(self):
        return '{0} {1}{2}{3}'.format(
            self.name,
            self.type.compile(),
            {False: ' not null'}.get(self.nullable, ''),
            {None: ''}.get(self.default, ' default {0}'.format(self.default)))

    def title(self):
        return self.name

    def subtitle(self):
        return self.table.title()

    def autocomplete(self):
        return '%s%s?%s' % (self.table.uri, self.table.name, self.name)

    def icon(self):
        return 'images/table.png'

    def format(self):
        return Formatter.format_column(self)

    def as_json(self):
        d = {
            '__cls__': str(self.__class__),
            'name': self.name,
            'table': self.table.name,
            'type': self.type.compile(),
            'uri': self.uri
        }
        if self.default is not None:
            d['default'] = self.default
        return d


def create_column(table, name, column=None):
    if column is None:
        return Column(table, name)
    return Column(table, name, **dictminus(column.__dict__, 'name', 'table'))



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import sys

from dbmanagr.writer import Writer

from dbmanagr.wrapper import Wrapper
from dbmanagr.config import Config
from dbmanagr.sources.source import Source
from dbmanagr.exception import UnknownTableException, \
    UnknownConnectionException
from dbmanagr.dto.mapper import to_dto

from .args import parser
from .writer import DiffWriter


def column_ddl(c):
    return c.ddl()


def column_name(c):
    return c.name


class DatabaseDiffer(Wrapper):
    """The main class"""
    def __init__(self, left, right):
        Wrapper.__init__(self, left)

        self.left = left
        self.right = right

        if left.formatter:
            Writer.set(left.formatter(left, right))
        else:
            Writer.set(DiffWriter(left, right))

    def execute(self):
        """The main method that splits the arguments and starts the magic"""
        left, right = self.left, self.right

        lcon = Source.connection(left)
        if not lcon:
            raise UnknownConnectionException(
                left.uri,
                map(lambda c: c.autocomplete(), Source.connections()))
        lopts = left.get(lcon.dbms)
        rcon = Source.connection(right)
        if not rcon:
            raise UnknownConnectionException(
                right.uri,
                map(lambda c: c.autocomplete(), Source.connections()))
        ropts = right.get(rcon.dbms)

        try:
            lcon.connect(lopts.database)
            rcon.connect(ropts.database)
            ltables = lcon.tables()
            if lopts.table not in ltables:
                raise UnknownTableException(lopts.table, ltables.keys())
            ltable = ltables[lopts.table]
            rtables = rcon.tables()
            if ropts.table not in rtables:
                raise UnknownTableException(ropts.table, rtables.keys())
            rtable = rtables[ropts.table]

            lcols = map(
                column_ddl if left.compare_ddl else column_name,
                ltable.columns())
            rcols = map(
                column_ddl if right.compare_ddl else column_name,
                rtable.columns())

            lplus = dict(map(
                lambda c: (c.split()[0], ltable.column(c.split()[0])),
                list(set(lcols) - set(rcols))))
            rplus = dict(map(
                lambda c: (c.split()[0], rtable.column(c.split()[0])),
                list(set(rcols) - set(lcols))))

            r = {}
            for k, v in lplus.iteritems():
                if k in rplus:
                    r[k] = (v, rplus[k])
                else:
                    r[k] = (v, None)
            for k, v in rplus.iteritems():
                if k not in lplus:
                    r[k] = (None, v)

            return to_dto(map(lambda (k, v): v, r.iteritems()))
        finally:
            lcon.close()
            rcon.close()


def init(argv, parser):
    left = Config.init(argv, parser)
    left.uri = left.left
    left.update_parsers()
    right = Config.init(argv, parser)
    right.uri = right.right
    right.update_parsers()
    return (left, right)


def execute(args):
    """
    Directly calls the execute method and avoids using the wrapper
    """
    return DatabaseDiffer(*init(args, parser)).execute()


def run(args):
    return DatabaseDiffer(*init(args, parser)).run()


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    return DatabaseDiffer(*init(args, parser)).write()



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

from dbmanagr.model.baseitem import BaseItem
from dbmanagr.formatter import Formatter
from dbmanagr.utils import primary_key_or_first_column

logger = logging.getLogger(__name__)


def val(row, column):
    # colname = '%s_title' % column
    # if colname in row.row:
    #     return '%s (%s)' % (row.row[colname], row.row[column])
    return row[column]


class Row(BaseItem):
    """A table row from the database"""

    def __init__(self, table, row):
        self.table = table
        self.row = row

    def __getitem__(self, i):
        if i is None:
            return None
        if type(i) == unicode:
            i = i.encode('ascii')
        if type(i) is str:
            try:
                return self.row.__dict__[i]
            except BaseException:
                return None
        return self.row[i]

    def __repr__(self):
        return str(self.row)

    def title(self):
        if 'title' in self.row.__dict__:
            return val(self, 'title')
        return val(self, primary_key_or_first_column(self.table))

    def subtitle(self):
        return val(self, 'subtitle')

    def autocomplete(self):
        column = primary_key_or_first_column(self.table)
        return self.table.autocomplete_(column, self[column])

    def icon(self):
        return 'images/row.png'

    def format(self):
        return Formatter.format_row(self)



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

from dbmanagr.formatter import Formatter
from dbmanagr.model import Model
from dbmanagr.logger import LogWith
from dbmanagr.utils import hash_

logger = logging.getLogger(__name__)


class BaseItem(Model):
    def title(self):  # pragma: no cover
        return 'Title'

    def subtitle(self):  # pragma: no cover
        return 'Subtitle'

    def autocomplete(self):  # pragma: no cover
        return 'Autocomplete'

    def validity(self):
        return True

    def icon(self):  # pragma: no cover
        return 'images/icon.png'

    def value(self):
        return self.title()

    @LogWith(logger)
    def uid(self):
        return hash_(self.autocomplete())

    def format(self):
        return Formatter.format(self)



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from dbmanagr.writer import FormatWriter
from dbmanagr.formatter import Formatter, DefaultFormatter

DEFAULT_FORMAT = u'{0}'
FORMATS = {
    -1: u'{0}\n',  # for testing only
    0: u'{0}\n',
    1: u'PID\tDatabase\tUser\tClient\tTX Start\tQuery Start\tState\t'
       u'Blocked by\tQuery\n{0}\n',
    2: u'PID\tDatabase\tUser\tClient\tTX Start\tTX Duration\tQuery Start\t'
       u'Query Duration\tState\tBlocked by\tQuery\n{0}\n'
}
ITEM_FORMATS = {
    -1: u'{row}',  # for testing only
    0: u'{row.pid}\t{row.state}\t{row.query_duration}\t{row.query}',
    1: u'{row.pid}\t{row.database_name}\t{row.username}\t{row.client}\t'
       u'{row.transaction_start:%Y-%m-%d %H:%M:%S}\t'
       u'{row.query_start:%Y-%m-%d %H:%M:%S}\t{row.state}\t'
       u'{row.blocked_by}\t{row.query}',
    2: u'{row.pid}\t{row.database_name}\t{row.username}\t{row.client}\t'
       u'{row.transaction_start}\t{row.transaction_duration}\t'
       u'{row.query_start}\t{row.query_duration}\t{row.state}\t'
       u'{row.blocked_by}\t{row.query}'
}


class StatementActivityWriter(FormatWriter):
    def __init__(self, options):
        FormatWriter.__init__(
            self,
            FORMATS.get(options.verbose, FORMATS[0]),
            ITEM_FORMATS.get(options.verbose, ITEM_FORMATS[0]))
        Formatter.set(DefaultFormatter())
        self.options = options

    def itemtostring(self, item):
        row = item.row
        return self.item_format.format(row=row)



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

from os.path import basename

from dbmanagr.logger import LogWith
from dbmanagr.model.databaseconnection import DatabaseConnection
from dbmanagr.model.database import Database

AUTOCOMPLETE_FORMAT = "{connection}/"

logger = logging.getLogger(__name__)


class SQLiteDatabase(Database):
    def __init__(self, connection):
        Database.__init__(self, connection, '', AUTOCOMPLETE_FORMAT)


class SQLiteConnection(DatabaseConnection):
    def __init__(self, uri, *args):
        self.path = args[2]
        self.filename = basename(self.path)
        DatabaseConnection.__init__(
            self,
            dbms='sqlite',
            database=self.databases()[0],
            uri=uri,
            subtitle='SQLite Connection')

    def __repr__(self):
        return AUTOCOMPLETE_FORMAT.format(connection=self.filename)

    def matches(self, options):
        if options.uri:
            return options.uri.startswith(self.filename)
        return False

    def filter_(self, options):
        options = options.get(self.dbms)
        return not options.uri or options.uri in self.filename

    @LogWith(logger)
    def connect(self, database=None):
        self.connect_to(self.uri.format(file=self.path))

    def databases(self):
        return [SQLiteDatabase(self)]



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from dbmanagr.jsonable import Jsonable, from_json
from dbmanagr import utils


def to_dto(model):
    if type(model) is dict:
        return dict(map(lambda (k, v): (k, to_dto(v)), model.iteritems()))
    if type(model) in (tuple, list, set):
        return map(to_dto, model)
    from dbmanagr.model.baseitem import BaseItem
    if isinstance(model, BaseItem):
        return Item(
            title=model.title(),
            subtitle=model.subtitle(),
            autocomplete=model.autocomplete(),
            validity=model.validity(),
            icon=model.icon(),
            value=model.value(),
            uid=model.uid(),
            format_=model.format()
        )
    return model


class Item(Jsonable):
    def __init__(
            self,
            title=None,
            subtitle=None,
            autocomplete=None,
            uid=None,
            icon=None,
            value=None,
            validity=None,
            format_=None):
        self.title_ = title
        self.subtitle_ = subtitle
        self.autocomplete_ = autocomplete
        self.uid_ = uid
        self.icon_ = icon
        self.value_ = value
        self.validity_ = validity
        self.format_ = format_

    def __hash__(self):
        return utils.hash_(utils.freeze(self.__dict__))

    def __eq__(self, o):
        return utils.hash_(self) == utils.hash_(o)

    def autocomplete(self):
        return self.autocomplete_

    def title(self):
        return self.title_

    def subtitle(self):
        return self.subtitle_

    def value(self):
        return self.value_

    def validity(self):
        return self.validity_

    def uid(self):
        if self.uid_ is not None:
            return self.uid_
        return utils.hash_(self.autocomplete())

    def icon(self):
        if self.icon_ is not None:
            return self.icon_
        return 'images/icon.png'

    def format(self):
        return self.format_

    @staticmethod
    def from_json(d):
        return Item(
            title=from_json(d['title']),
            subtitle=from_json(d['subtitle']),
            autocomplete=from_json(d['autocomplete']),
            uid=from_json(d['uid']),
            icon=from_json(d['icon']),
            value=from_json(d['value']),
            validity=from_json(d['validity']),
            format_=from_json(d['format_'])
        )



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

from dbmanagr.dto.row import Row
from dbmanagr.dto.column import Column
from dbmanagr.dto.foreignkey import ForeignKey
from dbmanagr.dto.table import Table
from dbmanagr.dto.node import ColumnNode, ForeignKeyNode, TableNode, NameNode

logger = logging.getLogger(__name__)


def to_dto(model):
    if type(model) is dict:
        return dict(map(lambda (k, v): (k, to_dto(v)), model.iteritems()))
    if type(model) in (tuple, list, set):
        return map(to_dto, model)
    if model.__class__.__name__ == 'Row':
        return Row(
            table=to_dto(model.table),
            row=to_dto(model.row),
            autocomplete=model.autocomplete()
        )
    if model.__class__.__name__ == 'Column':
        return Column(
            name=model.name,
            tablename=model.table.name,
            type_=model.type,
            nullable=model.nullable,
            default=model.default,
            autoincrement=model.autoincrement,
            primary_key=model.primary_key,
            autocomplete=model.autocomplete()
        )
    if model.__class__.__name__ == 'ForeignKey':
        return ForeignKey(a=to_dto(model.a), b=to_dto(model.b))
    if model.__class__.__name__ == 'Table':
        return Table(
            name=model.name,
            uri=model.uri,
            owner=model.owner,
            size=model.size,
            primary_key=model.primary_key,
            columns=to_dto(model.columns()),
            foreign_keys=to_dto(model.foreign_keys()),
            autocomplete=model.autocomplete()
        )
    if model.__class__.__name__ == 'ColumnNode':
        return ColumnNode(to_dto(model.column), model.indent)
    if model.__class__.__name__ == 'ForeignKeyNode':
        return ForeignKeyNode(
            to_dto(model.fk), to_dto(model.parent), model.indent)
    if model.__class__.__name__ == 'TableNode':
        return TableNode(to_dto(model.table), model.indent)
    if model.__class__.__name__ == 'NameNode':
        return NameNode(to_dto(model.name), model.indent)
    return model



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from dbmanagr.args import parent_parser, format_group, create_parser

from .writer import DiffWriter, DiffColumnWriter, DiffTestWriter

parent = parent_parser(daemonable=True)

group = format_group(parent, DiffTestWriter)
group.add_argument(
    '-D',
    '--default',
    default=True,
    help='output format: human readable hierarchical text',
    dest='formatter',
    action='store_const',
    const=DiffWriter)
group.add_argument(
    '-S',
    '--side-by-side',
    help='output format: compare side-by-side in two columns',
    dest='formatter',
    action='store_const',
    const=DiffColumnWriter)

parser = create_parser(
    prog='dbdiff',
    description='A diff tool that compares the structure of two database '
                'tables with each other.',
    parents=[parent])
parser.add_argument(
    'left',
    help='the left URI to parse (format for PostgreSQL/MySQL: '
         'user@host/database/table; for SQLite: databasefile.db/table)')
parser.add_argument(
    'right',
    help='the right URI to parse (format for PostgreSQL/MySQL: '
         'user@host/database/table; for SQLite: databasefile.db/table)')
parser.add_argument(
    '-v',
    '--verbose',
    action='count',
    help='specify the verbosity of the output, increase the number of '
         'occurences of this option to increase verbosity')
parser.add_argument(
    '-c',
    '--compare-ddl',
    default=False,
    action='store_true',
    help='compares the DDLs for each column')
# parser.add_argument('-I', '--ignore-case', help='ignore case differences in '
#                     'table columns')
# parser.add_argument('-q', '--brief', help='output only whether files differ')
# parser.add_argument('-y', '--side-by-side', help='output in two columns')
# parser.add_argument('-r', '--recursive', help='recursively compare any '
#                     'subdirectories found')
# parser.add_argument('-i', '--include', help='include the specified columns '
#                     'and their foreign rows, if any (multiple columns can '
#                     'be specified by separating them with a comma)')
# parser.add_argument('-x', '--exclude', help='Exclude the specified columns')
# parser.add_argument('-s', '--report-identical-tables', help='report when '
#                     'two tables are the same')



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import logging

from dbmanagr.wrapper import Wrapper
from dbmanagr.config import Config
from dbmanagr.sources.source import Source
from dbmanagr.writer import Writer
from dbmanagr.dto.mapper import to_dto
from dbmanagr.jsonable import Jsonable, from_json
from dbmanagr.utils.sql import sanitise

from .args import parser
from .writer import StatementActivityWriter

logger = logging.getLogger(__name__)


class RowItem(Jsonable):
    def __init__(self, row):
        self.row = row

    def __hash__(self):
        return hash(self.row.row)

    def __eq__(self, o):
        return hash(self) == hash(o)

    @staticmethod
    def from_json(d):
        return RowItem(
            from_json(d['row'])
        )


class DatabaseStatus(Wrapper):
    def __init__(self, options):
        Wrapper.__init__(self, options)

        if options.formatter:
            Writer.set(options.formatter(options))
        else:
            Writer.set(StatementActivityWriter(options))

    def execute(self):
        options = self.options

        options.pattern = sanitise(options.pattern)

        cons = Source.connections()

        # search exact match of connection
        for connection in cons:
            opts = options.get(connection.dbms)
            if opts.show_code > 1 and connection.matches(opts):
                try:
                    connection.connect(opts.database)
                    return to_dto(map(
                        RowItem, opts.statement_activity(connection)))
                finally:
                    connection.close()

        raise Exception('Specify the complete URI of the connection')


def execute(args):
    """
    Directly calls the execute method and avoids using the wrapper
    """
    return DatabaseStatus(Config.init(args, parser)).execute()


def run(args):
    return DatabaseStatus(Config.init(args, parser)).run()


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    return DatabaseStatus(Config.init(args, parser)).write()



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

logger = logging.getLogger(__name__)


class Source(object):
    sources = []

    def __init__(self):
        self._connections = []

    def list(self):
        return self._connections

    @staticmethod
    def connections():
        cons = []
        for source in Source.sources:
            cons += source.list()
        return set(cons)

    @staticmethod
    def connection(options):
        # search exact match of connection
        for connection in Source.connections():
            opts = options.get(connection.dbms)
            if connection.matches(opts):
                return connection
        return None
